# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2024 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import Awaitable, Callable
import asyncio
import logging


class BasicScheduler:
    background_loop: asyncio.Task | None
    tasks: set[asyncio.Task]
    log: logging.Logger

    def __init__(self, log: logging.Logger) -> None:
        self.log = log
        self.tasks = set()

    def _find_caller(self) -> str:
        try:
            file_name, line_number, function_name, _ = self.log.findCaller()
            return f"{function_name} at {file_name}:{line_number}"
        except ValueError:
            return "unknown function"

    def run_periodically(
        self,
        period: float | int,
        func: Callable[[], Awaitable],
        run_task_in_background: bool = False,
        catch_errors: bool = True,
    ) -> asyncio.Task:
        """
        Run a function periodically in the background.

        Args:
            period: The period in seconds between each call to the function.
            func: The function to run. No parameters will be provided,
                use :meth:`functools.partial` if you need to pass parameters.
            run_task_in_background: If ``True``, the function will be run in a background task.
                If ``False`` (the default), the loop will wait for the task to return before
                sleeping for the next period.
            catch_errors: Whether the scheduler should catch and log any errors.
                If ``False``, errors will be raised, and the caller must await the returned task
                to find errors. This parameter has no effect if ``run_task_in_background``
                is ``True``.

        Returns:
            The asyncio task object representing the background loop.
        """
        task = asyncio.create_task(
            self._call_periodically(
                period,
                func,
                caller=self._find_caller(),
                catch_errors=catch_errors,
                run_task_in_background=run_task_in_background,
            )
        )
        self._register_task(task)
        return task

    def run_later(
        self, delay: float | int, coro: Awaitable, catch_errors: bool = True
    ) -> asyncio.Task:
        """
        Run a coroutine after a delay.

        Examples:
            >>> self.sched.run_later(5, self.async_task(meow=True))

        Args:
            delay: The delay in seconds to await the coroutine after.
            coro: The coroutine to await.
            catch_errors: Whether the scheduler should catch and log any errors.
                If ``False``, errors will be raised, and the caller must await the returned task
                to find errors.

        Returns:
            The asyncio task object representing the scheduled task.
        """
        task = asyncio.create_task(
            self._call_with_delay(
                delay, coro, caller=self._find_caller(), catch_errors=catch_errors
            )
        )
        self._register_task(task)
        return task

    def _register_task(self, task: asyncio.Task) -> None:
        self.tasks.add(task)
        task.add_done_callback(self.tasks.discard)

    async def _call_periodically(
        self,
        period: float | int,
        func: Callable[[], Awaitable],
        caller: str,
        catch_errors: bool,
        run_task_in_background: bool,
    ) -> None:
        while True:
            try:
                await asyncio.sleep(period)
                if run_task_in_background:
                    self._register_task(
                        asyncio.create_task(self._call_periodically_background(func(), caller))
                    )
                else:
                    await func()
            except asyncio.CancelledError:
                raise
            except Exception:
                if catch_errors:
                    self.log.exception(f"Uncaught error in background loop (created in {caller})")
                else:
                    raise

    async def _call_periodically_background(self, coro: Awaitable, caller: str) -> None:
        try:
            await coro
        except asyncio.CancelledError:
            raise
        except Exception:
            self.log.exception(f"Uncaught error in background loop subtask (created in {caller})")

    async def _call_with_delay(
        self, delay: float | int, coro: Awaitable, caller: str, catch_errors: bool
    ) -> None:
        try:
            await asyncio.sleep(delay)
            await coro
        except asyncio.CancelledError:
            raise
        except Exception:
            if catch_errors:
                self.log.exception(f"Uncaught error in scheduled task (created in {caller})")
            else:
                raise

    def stop(self) -> None:
        """
        Stop all scheduled tasks and background loops.
        """
        for task in self.tasks:
            task.cancel(msg="Scheduler stopped")



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Any, Awaitable, Callable

from aiohttp import hdrs, web

WebHandler = Callable[[web.Request], Awaitable[web.StreamResponse]]
WebHandlerDecorator = Callable[[WebHandler], WebHandler]


def head(path: str, **kwargs: Any) -> WebHandlerDecorator:
    return handle(hdrs.METH_HEAD, path, **kwargs)


def options(path: str, **kwargs: Any) -> WebHandlerDecorator:
    return handle(hdrs.METH_OPTIONS, path, **kwargs)


def get(path: str, **kwargs: Any) -> WebHandlerDecorator:
    return handle(hdrs.METH_GET, path, **kwargs)


def post(path: str, **kwargs: Any) -> WebHandlerDecorator:
    return handle(hdrs.METH_POST, path, **kwargs)


def put(path: str, **kwargs: Any) -> WebHandlerDecorator:
    return handle(hdrs.METH_PUT, path, **kwargs)


def patch(path: str, **kwargs: Any) -> WebHandlerDecorator:
    return handle(hdrs.METH_PATCH, path, **kwargs)


def delete(path: str, **kwargs: Any) -> WebHandlerDecorator:
    return handle(hdrs.METH_DELETE, path, **kwargs)


def view(path: str, **kwargs: Any) -> WebHandlerDecorator:
    return handle(hdrs.METH_ANY, path, **kwargs)


def handle(method: str, path: str, **kwargs) -> WebHandlerDecorator:
    def decorator(handler: WebHandler) -> WebHandler:
        try:
            handlers = getattr(handler, "__mb_web_handler__")
        except AttributeError:
            handlers = []
            setattr(handler, "__mb_web_handler__", handlers)
        handlers.append((method, path, kwargs))
        return handler

    return decorator



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Iterable,
    List,
    NewType,
    Optional,
    Pattern,
    Sequence,
    Set,
    Tuple,
    Union,
)
from abc import ABC, abstractmethod
import asyncio
import functools
import inspect
import re

from mautrix.types import EventType, MessageType

from ..matrix import MaubotMessageEvent
from . import event

PrefixType = Optional[Union[str, Callable[[], str], Callable[[Any], str]]]
AliasesType = Union[
    List[str], Tuple[str, ...], Set[str], Callable[[str], bool], Callable[[Any, str], bool]
]
CommandHandlerFunc = NewType(
    "CommandHandlerFunc", Callable[[MaubotMessageEvent, Any], Awaitable[Any]]
)
CommandHandlerDecorator = NewType(
    "CommandHandlerDecorator",
    Callable[[Union["CommandHandler", CommandHandlerFunc]], "CommandHandler"],
)
PassiveCommandHandlerDecorator = NewType(
    "PassiveCommandHandlerDecorator", Callable[[CommandHandlerFunc], CommandHandlerFunc]
)


def _split_in_two(val: str, split_by: str) -> List[str]:
    return val.split(split_by, 1) if split_by in val else [val, ""]


class CommandHandler:
    def __init__(self, func: CommandHandlerFunc) -> None:
        self.__mb_func__: CommandHandlerFunc = func
        self.__mb_parent__: Optional[CommandHandler] = None
        self.__mb_subcommands__: List[CommandHandler] = []
        self.__mb_arguments__: List[Argument] = []
        self.__mb_help__: Optional[str] = None
        self.__mb_get_name__: Callable[[Any], str] = lambda s: "noname"
        self.__mb_is_command_match__: Callable[[Any, str], bool] = self.__command_match_unset
        self.__mb_require_subcommand__: bool = True
        self.__mb_must_consume_args__: bool = True
        self.__mb_arg_fallthrough__: bool = True
        self.__mb_event_handler__: bool = True
        self.__mb_event_types__: set[EventType] = {EventType.ROOM_MESSAGE}
        self.__mb_msgtypes__: Iterable[MessageType] = (MessageType.TEXT,)
        self.__bound_copies__: Dict[Any, CommandHandler] = {}
        self.__bound_instance__: Any = None

    def __get__(self, instance, instancetype):
        if not instance or self.__bound_instance__:
            return self
        try:
            return self.__bound_copies__[instance]
        except KeyError:
            new_ch = type(self)(self.__mb_func__)
            keys = [
                "parent",
                "subcommands",
                "arguments",
                "help",
                "get_name",
                "is_command_match",
                "require_subcommand",
                "must_consume_args",
                "arg_fallthrough",
                "event_handler",
                "event_types",
                "msgtypes",
            ]
            for key in keys:
                key = f"__mb_{key}__"
                setattr(new_ch, key, getattr(self, key))
            new_ch.__bound_instance__ = instance
            new_ch.__mb_subcommands__ = [
                subcmd.__get__(instance, instancetype) for subcmd in self.__mb_subcommands__
            ]
            self.__bound_copies__[instance] = new_ch
            return new_ch

    @staticmethod
    def __command_match_unset(self, val: str) -> bool:
        raise NotImplementedError("Hmm")

    async def __call__(
        self,
        evt: MaubotMessageEvent,
        *,
        _existing_args: Dict[str, Any] = None,
        remaining_val: str = None,
    ) -> Any:
        if evt.sender == evt.client.mxid or evt.content.msgtype not in self.__mb_msgtypes__:
            return
        if remaining_val is None:
            if not evt.content.body or evt.content.body[0] != "!":
                return
            command, remaining_val = _split_in_two(evt.content.body[1:], " ")
            command = command.lower()
            if not self.__mb_is_command_match__(self.__bound_instance__, command):
                return
        call_args: Dict[str, Any] = {**_existing_args} if _existing_args else {}

        if not self.__mb_arg_fallthrough__ and len(self.__mb_subcommands__) > 0:
            ok, res = await self.__call_subcommand__(evt, call_args, remaining_val)
            if ok:
                return res

        ok, remaining_val = await self.__parse_args__(evt, call_args, remaining_val)
        if not ok:
            return
        elif self.__mb_arg_fallthrough__ and len(self.__mb_subcommands__) > 0:
            ok, res = await self.__call_subcommand__(evt, call_args, remaining_val)
            if ok:
                return res
            elif self.__mb_require_subcommand__:
                await evt.reply(self.__mb_full_help__)
                return

        if self.__mb_must_consume_args__ and remaining_val.strip():
            await evt.reply(self.__mb_full_help__)
            return

        if self.__bound_instance__:
            return await self.__mb_func__(self.__bound_instance__, evt, **call_args)
        return await self.__mb_func__(evt, **call_args)

    async def __call_subcommand__(
        self, evt: MaubotMessageEvent, call_args: Dict[str, Any], remaining_val: str
    ) -> Tuple[bool, Any]:
        command, remaining_val = _split_in_two(remaining_val.strip(), " ")
        for subcommand in self.__mb_subcommands__:
            if subcommand.__mb_is_command_match__(subcommand.__bound_instance__, command):
                return True, await subcommand(
                    evt, _existing_args=call_args, remaining_val=remaining_val
                )
        return False, None

    async def __parse_args__(
        self, evt: MaubotMessageEvent, call_args: Dict[str, Any], remaining_val: str
    ) -> Tuple[bool, str]:
        for arg in self.__mb_arguments__:
            try:
                remaining_val, call_args[arg.name] = arg.match(
                    remaining_val.strip(), evt=evt, instance=self.__bound_instance__
                )
                if arg.required and call_args[arg.name] is None:
                    raise ValueError("Argument required")
            except ArgumentSyntaxError as e:
                await evt.reply(e.message + (f"\n{self.__mb_usage__}" if e.show_usage else ""))
                return False, remaining_val
            except ValueError:
                await evt.reply(self.__mb_usage__)
                return False, remaining_val
        return True, remaining_val

    @property
    def __mb_full_help__(self) -> str:
        usage = self.__mb_usage_without_subcommands__ + "\n\n"
        if not self.__mb_require_subcommand__:
            usage += f"* {self.__mb_prefix__} {self.__mb_usage_args__} - {self.__mb_help__}\n"
        usage += "\n".join(cmd.__mb_usage_inline__ for cmd in self.__mb_subcommands__)
        return usage

    @property
    def __mb_usage_args__(self) -> str:
        arg_usage = " ".join(
            f"<{arg.label}>" if arg.required else f"[{arg.label}]" for arg in self.__mb_arguments__
        )
        if self.__mb_subcommands__ and self.__mb_arg_fallthrough__:
            arg_usage += " " + self.__mb_usage_subcommand__
        return arg_usage

    @property
    def __mb_usage_subcommand__(self) -> str:
        return f"<subcommand> [...]"

    @property
    def __mb_name__(self) -> str:
        return self.__mb_get_name__(self.__bound_instance__)

    @property
    def __mb_prefix__(self) -> str:
        if self.__mb_parent__:
            return (
                f"!{self.__mb_parent__.__mb_get_name__(self.__bound_instance__)} "
                f"{self.__mb_name__}"
            )
        return f"!{self.__mb_name__}"

    @property
    def __mb_usage_inline__(self) -> str:
        if not self.__mb_arg_fallthrough__:
            return (
                f"* {self.__mb_name__} {self.__mb_usage_args__} - {self.__mb_help__}\n"
                f"* {self.__mb_name__} {self.__mb_usage_subcommand__}"
            )
        return f"* {self.__mb_name__} {self.__mb_usage_args__} - {self.__mb_help__}"

    @property
    def __mb_subcommands_list__(self) -> str:
        return f"**Subcommands:** {', '.join(sc.__mb_name__ for sc in self.__mb_subcommands__)}"

    @property
    def __mb_usage_without_subcommands__(self) -> str:
        if not self.__mb_arg_fallthrough__:
            if not self.__mb_arguments__:
                return f"**Usage:** {self.__mb_prefix__} [subcommand] [...]"
            return (
                f"**Usage:** {self.__mb_prefix__} {self.__mb_usage_args__}"
                f" _OR_ {self.__mb_prefix__} {self.__mb_usage_subcommand__}"
            )
        return f"**Usage:** {self.__mb_prefix__} {self.__mb_usage_args__}"

    @property
    def __mb_usage__(self) -> str:
        if len(self.__mb_subcommands__) > 0:
            return f"{self.__mb_usage_without_subcommands__}  \n{self.__mb_subcommands_list__}"
        return self.__mb_usage_without_subcommands__

    def subcommand(
        self,
        name: PrefixType = None,
        *,
        help: str = None,
        aliases: AliasesType = None,
        required_subcommand: bool = True,
        arg_fallthrough: bool = True,
    ) -> CommandHandlerDecorator:
        def decorator(func: Union[CommandHandler, CommandHandlerFunc]) -> CommandHandler:
            if not isinstance(func, CommandHandler):
                func = CommandHandler(func)
            new(
                name,
                help=help,
                aliases=aliases,
                require_subcommand=required_subcommand,
                arg_fallthrough=arg_fallthrough,
            )(func)
            func.__mb_parent__ = self
            func.__mb_event_handler__ = False
            self.__mb_subcommands__.append(func)
            return func

        return decorator


def new(
    name: PrefixType = None,
    *,
    help: str = None,
    aliases: AliasesType = None,
    event_type: EventType = EventType.ROOM_MESSAGE,
    msgtypes: Iterable[MessageType] = None,
    require_subcommand: bool = True,
    arg_fallthrough: bool = True,
    must_consume_args: bool = True,
) -> CommandHandlerDecorator:
    def decorator(func: Union[CommandHandler, CommandHandlerFunc]) -> CommandHandler:
        if not isinstance(func, CommandHandler):
            func = CommandHandler(func)
        func.__mb_help__ = help
        if name:
            if callable(name):
                if len(inspect.getfullargspec(name).args) == 0:
                    func.__mb_get_name__ = lambda self: name()
                else:
                    func.__mb_get_name__ = name
            else:
                func.__mb_get_name__ = lambda self: name
        else:
            func.__mb_get_name__ = lambda self: func.__mb_func__.__name__.replace("_", "-")
        if callable(aliases):
            if len(inspect.getfullargspec(aliases).args) == 1:
                func.__mb_is_command_match__ = lambda self, val: aliases(val)
            else:
                func.__mb_is_command_match__ = aliases
        elif isinstance(aliases, (list, set, tuple)):
            func.__mb_is_command_match__ = lambda self, val: (
                val == func.__mb_get_name__(self) or val in aliases
            )
        else:
            func.__mb_is_command_match__ = lambda self, val: val == func.__mb_get_name__(self)
        # Decorators are executed last to first, so we reverse the argument list.
        func.__mb_arguments__.reverse()
        func.__mb_require_subcommand__ = require_subcommand
        func.__mb_arg_fallthrough__ = arg_fallthrough
        func.__mb_must_consume_args__ = must_consume_args
        func.__mb_event_types__ = {event_type}
        if msgtypes:
            func.__mb_msgtypes__ = msgtypes
        return func

    return decorator


class ArgumentSyntaxError(ValueError):
    def __init__(self, message: str, show_usage: bool = True) -> None:
        super().__init__(message)
        self.message = message
        self.show_usage = show_usage


class Argument(ABC):
    def __init__(
        self, name: str, label: str = None, *, required: bool = False, pass_raw: bool = False
    ) -> None:
        self.name = name
        self.label = label or name
        self.required = required
        self.pass_raw = pass_raw

    @abstractmethod
    def match(self, val: str, **kwargs) -> Tuple[str, Any]:
        pass

    def __call__(self, func: Union[CommandHandler, CommandHandlerFunc]) -> CommandHandler:
        if not isinstance(func, CommandHandler):
            func = CommandHandler(func)
        func.__mb_arguments__.append(self)
        return func


class RegexArgument(Argument):
    def __init__(
        self,
        name: str,
        label: str = None,
        *,
        required: bool = False,
        pass_raw: bool = False,
        matches: str = None,
    ) -> None:
        super().__init__(name, label, required=required, pass_raw=pass_raw)
        matches = f"^{matches}" if self.pass_raw else f"^{matches}$"
        self.regex = re.compile(matches)

    def match(self, val: str, **kwargs) -> Tuple[str, Any]:
        orig_val = val
        if not self.pass_raw:
            val = re.split(r"\s", val, 1)[0]
        match = self.regex.match(val)
        if match:
            return (
                orig_val[: match.start()] + orig_val[match.end() :],
                match.groups() or val[match.start() : match.end()],
            )
        return orig_val, None


class CustomArgument(Argument):
    def __init__(
        self,
        name: str,
        label: str = None,
        *,
        required: bool = False,
        pass_raw: bool = False,
        matcher: Callable[[str], Any],
    ) -> None:
        super().__init__(name, label, required=required, pass_raw=pass_raw)
        self.matcher = matcher

    def match(self, val: str, **kwargs) -> Tuple[str, Any]:
        if self.pass_raw:
            return self.matcher(val)
        orig_val = val
        val = re.split(r"\s", val, 1)[0]
        res = self.matcher(val)
        if res is not None:
            return orig_val[len(val) :], res
        return orig_val, None


class SimpleArgument(Argument):
    def match(self, val: str, **kwargs) -> Tuple[str, Any]:
        if self.pass_raw:
            return "", val
        res = re.split(r"\s", val, 1)[0]
        return val[len(res) :], res


def argument(
    name: str,
    label: str = None,
    *,
    required: bool = True,
    matches: Optional[str] = None,
    parser: Optional[Callable[[str], Any]] = None,
    pass_raw: bool = False,
) -> CommandHandlerDecorator:
    if matches:
        return RegexArgument(name, label, required=required, matches=matches, pass_raw=pass_raw)
    elif parser:
        return CustomArgument(name, label, required=required, matcher=parser, pass_raw=pass_raw)
    else:
        return SimpleArgument(name, label, required=required, pass_raw=pass_raw)


def passive(
    regex: Union[str, Pattern],
    *,
    msgtypes: Sequence[MessageType] = (MessageType.TEXT,),
    field: Callable[[MaubotMessageEvent], str] = lambda evt: evt.content.body,
    event_type: EventType = EventType.ROOM_MESSAGE,
    multiple: bool = False,
    case_insensitive: bool = False,
    multiline: bool = False,
    dot_all: bool = False,
) -> PassiveCommandHandlerDecorator:
    if not isinstance(regex, Pattern):
        flags = re.RegexFlag.UNICODE
        if case_insensitive:
            flags |= re.IGNORECASE
        if multiline:
            flags |= re.MULTILINE
        if dot_all:
            flags |= re.DOTALL
        regex = re.compile(regex, flags=flags)

    def decorator(func: CommandHandlerFunc) -> CommandHandlerFunc:
        combine = None
        if hasattr(func, "__mb_passive_orig__"):
            combine = func
            func = func.__mb_passive_orig__

        @event.on(event_type)
        @functools.wraps(func)
        async def replacement(self, evt: MaubotMessageEvent = None) -> None:
            if not evt and isinstance(self, MaubotMessageEvent):
                evt = self
                self = None
            if evt.sender == evt.client.mxid:
                return
            elif msgtypes and evt.content.msgtype not in msgtypes:
                return
            data = field(evt)
            if multiple:
                val = [
                    (data[match.pos : match.endpos], *match.groups())
                    for match in regex.finditer(data)
                ]
            else:
                match = regex.search(data)
                if match:
                    val = (data[match.pos : match.endpos], *match.groups())
                else:
                    val = None
            if val:
                if self:
                    await func(self, evt, val)
                else:
                    await func(evt, val)

        if combine:
            orig_replacement = replacement

            @event.on(event_type)
            @functools.wraps(func)
            async def replacement(self, evt: MaubotMessageEvent = None) -> None:
                await asyncio.gather(combine(self, evt), orig_replacement(self, evt))

        replacement.__mb_passive_orig__ = func

        return replacement

    return decorator



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2023 Aurélien Bompard
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from pathlib import Path
import asyncio
import logging

from ruamel.yaml import YAML
import aiohttp
import pytest
import pytest_asyncio

from maubot import Plugin
from maubot.loader import PluginMeta
from maubot.standalone.loader import FileSystemLoader
from mautrix.util.async_db import Database
from mautrix.util.config import BaseProxyConfig, RecursiveDict
from mautrix.util.logging import TraceLogger

from .bot import TestBot


@pytest_asyncio.fixture
async def maubot_test_bot():
    return TestBot()


@pytest.fixture
def maubot_upgrade_table():
    return None


@pytest.fixture
def maubot_plugin_path():
    return Path(".")


@pytest.fixture
def maubot_plugin_meta(maubot_plugin_path):
    yaml = YAML()
    with open(maubot_plugin_path.joinpath("maubot.yaml")) as fh:
        plugin_meta = PluginMeta.deserialize(yaml.load(fh.read()))
    return plugin_meta


@pytest_asyncio.fixture
async def maubot_plugin_db(tmp_path, maubot_plugin_meta, maubot_upgrade_table):
    if not maubot_plugin_meta.get("database", False):
        return
    db_path = tmp_path.joinpath("maubot-tests.db").as_posix()
    db = Database.create(
        f"sqlite:{db_path}",
        upgrade_table=maubot_upgrade_table,
        log=logging.getLogger("db"),
    )
    await db.start()
    yield db
    await db.stop()


@pytest.fixture
def maubot_plugin_class():
    return Plugin


@pytest.fixture
def maubot_plugin_config_class():
    return BaseProxyConfig


@pytest.fixture
def maubot_plugin_config_dict():
    return {}


@pytest.fixture
def maubot_plugin_config_overrides():
    return {}


@pytest.fixture
def maubot_plugin_config(
    maubot_plugin_path,
    maubot_plugin_config_class,
    maubot_plugin_config_dict,
    maubot_plugin_config_overrides,
):
    yaml = YAML()
    with open(maubot_plugin_path.joinpath("base-config.yaml")) as fh:
        base_config = RecursiveDict(yaml.load(fh))
    maubot_plugin_config_dict.update(maubot_plugin_config_overrides)
    return maubot_plugin_config_class(
        load=lambda: maubot_plugin_config_dict,
        load_base=lambda: base_config,
        save=lambda c: None,
    )


@pytest_asyncio.fixture
async def maubot_plugin(
    maubot_test_bot,
    maubot_plugin_db,
    maubot_plugin_class,
    maubot_plugin_path,
    maubot_plugin_config,
    maubot_plugin_meta,
):
    loader = FileSystemLoader(maubot_plugin_path, maubot_plugin_meta)
    async with aiohttp.ClientSession() as http:
        instance = maubot_plugin_class(
            client=maubot_test_bot.client,
            loop=asyncio.get_running_loop(),
            http=http,
            instance_id="tests",
            log=TraceLogger("test"),
            config=maubot_plugin_config,
            database=maubot_plugin_db,
            webapp=None,
            webapp_url=None,
            loader=loader,
        )
        await instance.internal_start()
        yield instance



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import json

from aiohttp import web

from .auth import create_token
from .base import get_config, routes
from .responses import resp


@routes.post("/auth/login")
async def login(request: web.Request) -> web.Response:
    try:
        data = await request.json()
    except json.JSONDecodeError:
        return resp.body_not_json
    secret = data.get("secret")
    if secret and get_config()["server.unshared_secret"] == secret:
        user = data.get("user") or "root"
        return resp.logged_in(create_token(user))

    username = data.get("username")
    password = data.get("password")
    if get_config().check_password(username, password):
        return resp.logged_in(create_token(username))

    return resp.bad_auth



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Callable
import os

from packaging.version import InvalidVersion, Version
from prompt_toolkit.document import Document
from prompt_toolkit.validation import ValidationError, Validator
import click

from ..util import spdx as spdxlib


class Required(Validator):
    proxy: Validator

    def __init__(self, proxy: Validator = None) -> None:
        self.proxy = proxy

    def validate(self, document: Document) -> None:
        if len(document.text) == 0:
            raise ValidationError(message="This field is required")
        if self.proxy:
            return self.proxy.validate(document)


class ClickValidator(Validator):
    click_type: Callable[[str], str] = None

    @classmethod
    def validate(cls, document: Document) -> None:
        try:
            cls.click_type(document.text)
        except click.BadParameter as e:
            raise ValidationError(message=e.message, cursor_position=len(document.text))


def path(val: str) -> str:
    val = os.path.abspath(val)
    if os.path.exists(val):
        return val
    directory = os.path.dirname(val)
    if not os.path.isdir(directory):
        if os.path.exists(directory):
            raise click.BadParameter(f"{directory} is not a directory")
        raise click.BadParameter(f"{directory} does not exist")
    return val


class PathValidator(ClickValidator):
    click_type = path


def version(val: str) -> Version:
    try:
        return Version(val)
    except InvalidVersion as e:
        raise click.BadParameter(f"{val} is not a valid PEP-440 version") from e


class VersionValidator(ClickValidator):
    click_type = version


def spdx(val: str) -> str:
    if not spdxlib.valid(val):
        raise click.BadParameter(f"{val} is not a valid SPDX license identifier")
    return val


class SPDXValidator(ClickValidator):
    click_type = spdx



try:
    from sqlalchemy import MetaData, asc, create_engine, desc
    from sqlalchemy.engine import Engine
    from sqlalchemy.exc import IntegrityError, OperationalError
except ImportError:

    class FakeError(Exception):
        pass

    class FakeType:
        def __init__(self, *args, **kwargs):
            raise Exception("SQLAlchemy is not installed")

    def create_engine(*args, **kwargs):
        raise Exception("SQLAlchemy is not installed")

    MetaData = Engine = FakeType
    IntegrityError = OperationalError = FakeError
    asc = desc = lambda a: a



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import Callable, NewType

from mautrix.client import EventHandler, InternalEventType
from mautrix.types import EventType

EventHandlerDecorator = NewType("EventHandlerDecorator", Callable[[EventHandler], EventHandler])


def on(var: EventType | InternalEventType | EventHandler) -> EventHandlerDecorator | EventHandler:
    def decorator(func: EventHandler) -> EventHandler:
        func.__mb_event_handler__ = True
        if isinstance(var, (EventType, InternalEventType)):
            if hasattr(func, "__mb_event_types__"):
                func.__mb_event_types__.add(var)
            else:
                func.__mb_event_types__ = {var}
        else:
            func.__mb_event_types__ = {EventType.ALL}

        return func

    return decorator if isinstance(var, (EventType, InternalEventType)) else decorator(var)


def off(func: EventHandler) -> EventHandler:
    func.__mb_event_handler__ = False
    return func



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import TYPE_CHECKING
from http import HTTPStatus

from aiohttp import web
from asyncpg import PostgresError
import aiosqlite

if TYPE_CHECKING:
    from sqlalchemy.exc import IntegrityError, OperationalError


class _Response:
    @property
    def body_not_json(self) -> web.Response:
        return web.json_response(
            {
                "error": "Request body is not JSON",
                "errcode": "body_not_json",
            },
            status=HTTPStatus.BAD_REQUEST,
        )

    @property
    def plugin_type_required(self) -> web.Response:
        return web.json_response(
            {
                "error": "Plugin type is required when creating plugin instances",
                "errcode": "plugin_type_required",
            },
            status=HTTPStatus.BAD_REQUEST,
        )

    @property
    def primary_user_required(self) -> web.Response:
        return web.json_response(
            {
                "error": "Primary user is required when creating plugin instances",
                "errcode": "primary_user_required",
            },
            status=HTTPStatus.BAD_REQUEST,
        )

    @property
    def bad_client_access_token(self) -> web.Response:
        return web.json_response(
            {
                "error": "Invalid access token",
                "errcode": "bad_client_access_token",
            },
            status=HTTPStatus.BAD_REQUEST,
        )

    @property
    def bad_client_access_details(self) -> web.Response:
        return web.json_response(
            {
                "error": "Invalid homeserver or access token",
                "errcode": "bad_client_access_details",
            },
            status=HTTPStatus.BAD_REQUEST,
        )

    @property
    def bad_client_connection_details(self) -> web.Response:
        return web.json_response(
            {
                "error": "Could not connect to homeserver",
                "errcode": "bad_client_connection_details",
            },
            status=HTTPStatus.BAD_REQUEST,
        )

    def mxid_mismatch(self, found: str) -> web.Response:
        return web.json_response(
            {
                "error": (
                    "The Matrix user ID of the client and the user ID of the access token don't "
                    f"match. Access token is for user {found}"
                ),
                "errcode": "mxid_mismatch",
            },
            status=HTTPStatus.BAD_REQUEST,
        )

    def device_id_mismatch(self, found: str) -> web.Response:
        return web.json_response(
            {
                "error": (
                    "The Matrix device ID of the client and the device ID of the access token "
                    f"don't match. Access token is for device {found}"
                ),
                "errcode": "mxid_mismatch",
            },
            status=HTTPStatus.BAD_REQUEST,
        )

    @property
    def pid_mismatch(self) -> web.Response:
        return web.json_response(
            {
                "error": "The ID in the path does not match the ID of the uploaded plugin",
                "errcode": "pid_mismatch",
            },
            status=HTTPStatus.BAD_REQUEST,
        )

    @property
    def username_or_password_missing(self) -> web.Response:
        return web.json_response(
            {
                "error": "Username or password missing",
                "errcode": "username_or_password_missing",
            },
            status=HTTPStatus.BAD_REQUEST,
        )

    @property
    def query_missing(self) -> web.Response:
        return web.json_response(
            {
                "error": "Query missing",
                "errcode": "query_missing",
            },
            status=HTTPStatus.BAD_REQUEST,
        )

    @staticmethod
    def sql_error(error: PostgresError | aiosqlite.Error, query: str) -> web.Response:
        return web.json_response(
            {
                "ok": False,
                "query": query,
                "error": str(error),
                "errcode": "sql_error",
            },
            status=HTTPStatus.BAD_REQUEST,
        )

    @staticmethod
    def sql_operational_error(error: OperationalError, query: str) -> web.Response:
        return web.json_response(
            {
                "ok": False,
                "query": query,
                "error": str(error.orig),
                "full_error": str(error),
                "errcode": "sql_operational_error",
            },
            status=HTTPStatus.BAD_REQUEST,
        )

    @staticmethod
    def sql_integrity_error(error: IntegrityError, query: str) -> web.Response:
        return web.json_response(
            {
                "ok": False,
                "query": query,
                "error": str(error.orig),
                "full_error": str(error),
                "errcode": "sql_integrity_error",
            },
            status=HTTPStatus.BAD_REQUEST,
        )

    @property
    def bad_auth(self) -> web.Response:
        return web.json_response(
            {
                "error": "Invalid username or password",
                "errcode": "invalid_auth",
            },
            status=HTTPStatus.UNAUTHORIZED,
        )

    @property
    def no_token(self) -> web.Response:
        return web.json_response(
            {
                "error": "Authorization token missing",
                "errcode": "auth_token_missing",
            },
            status=HTTPStatus.UNAUTHORIZED,
        )

    @property
    def invalid_token(self) -> web.Response:
        return web.json_response(
            {
                "error": "Invalid authorization token",
                "errcode": "auth_token_invalid",
            },
            status=HTTPStatus.UNAUTHORIZED,
        )

    @property
    def plugin_not_found(self) -> web.Response:
        return web.json_response(
            {
                "error": "Plugin not found",
                "errcode": "plugin_not_found",
            },
            status=HTTPStatus.NOT_FOUND,
        )

    @property
    def client_not_found(self) -> web.Response:
        return web.json_response(
            {
                "error": "Client not found",
                "errcode": "client_not_found",
            },
            status=HTTPStatus.NOT_FOUND,
        )

    @property
    def primary_user_not_found(self) -> web.Response:
        return web.json_response(
            {
                "error": "Client for given primary user not found",
                "errcode": "primary_user_not_found",
            },
            status=HTTPStatus.NOT_FOUND,
        )

    @property
    def instance_not_found(self) -> web.Response:
        return web.json_response(
            {
                "error": "Plugin instance not found",
                "errcode": "instance_not_found",
            },
            status=HTTPStatus.NOT_FOUND,
        )

    @property
    def plugin_type_not_found(self) -> web.Response:
        return web.json_response(
            {
                "error": "Given plugin type not found",
                "errcode": "plugin_type_not_found",
            },
            status=HTTPStatus.NOT_FOUND,
        )

    @property
    def path_not_found(self) -> web.Response:
        return web.json_response(
            {
                "error": "Resource not found",
                "errcode": "resource_not_found",
            },
            status=HTTPStatus.NOT_FOUND,
        )

    @property
    def server_not_found(self) -> web.Response:
        return web.json_response(
            {
                "error": "Registration target server not found",
                "errcode": "server_not_found",
            },
            status=HTTPStatus.NOT_FOUND,
        )

    @property
    def registration_secret_not_found(self) -> web.Response:
        return web.json_response(
            {
                "error": "Config does not have a registration secret for that server",
                "errcode": "registration_secret_not_found",
            },
            status=HTTPStatus.NOT_FOUND,
        )

    @property
    def registration_no_sso(self) -> web.Response:
        return web.json_response(
            {
                "error": "The register operation is only for registering with a password",
                "errcode": "registration_no_sso",
            },
            status=HTTPStatus.BAD_REQUEST,
        )

    @property
    def sso_not_supported(self) -> web.Response:
        return web.json_response(
            {
                "error": "That server does not seem to support single sign-on",
                "errcode": "sso_not_supported",
            },
            status=HTTPStatus.FORBIDDEN,
        )

    @property
    def plugin_has_no_database(self) -> web.Response:
        return web.json_response(
            {
                "error": "Given plugin does not have a database",
                "errcode": "plugin_has_no_database",
            }
        )

    @property
    def unsupported_plugin_database(self) -> web.Response:
        return web.json_response(
            {
                "error": "The database type is not supported by this API",
                "errcode": "unsupported_plugin_database",
            }
        )

    @property
    def sqlalchemy_not_installed(self) -> web.Response:
        return web.json_response(
            {
                "error": "This plugin requires a legacy database, but SQLAlchemy is not installed",
                "errcode": "unsupported_plugin_database",
            },
            status=HTTPStatus.NOT_IMPLEMENTED,
        )

    @property
    def table_not_found(self) -> web.Response:
        return web.json_response(
            {
                "error": "Given table not found in plugin database",
                "errcode": "table_not_found",
            }
        )

    @property
    def method_not_allowed(self) -> web.Response:
        return web.json_response(
            {
                "error": "Method not allowed",
                "errcode": "method_not_allowed",
            },
            status=HTTPStatus.METHOD_NOT_ALLOWED,
        )

    @property
    def user_exists(self) -> web.Response:
        return web.json_response(
            {
                "error": "There is already a client with the user ID of that token",
                "errcode": "user_exists",
            },
            status=HTTPStatus.CONFLICT,
        )

    @property
    def plugin_exists(self) -> web.Response:
        return web.json_response(
            {
                "error": "A plugin with the same ID as the uploaded plugin already exists",
                "errcode": "plugin_exists",
            },
            status=HTTPStatus.CONFLICT,
        )

    @property
    def plugin_in_use(self) -> web.Response:
        return web.json_response(
            {
                "error": "Plugin instances of this type still exist",
                "errcode": "plugin_in_use",
            },
            status=HTTPStatus.PRECONDITION_FAILED,
        )

    @property
    def client_in_use(self) -> web.Response:
        return web.json_response(
            {
                "error": "Plugin instances with this client as their primary user still exist",
                "errcode": "client_in_use",
            },
            status=HTTPStatus.PRECONDITION_FAILED,
        )

    @staticmethod
    def plugin_import_error(error: str, stacktrace: str) -> web.Response:
        return web.json_response(
            {
                "error": error,
                "stacktrace": stacktrace,
                "errcode": "plugin_invalid",
            },
            status=HTTPStatus.BAD_REQUEST,
        )

    @staticmethod
    def plugin_reload_error(error: str, stacktrace: str) -> web.Response:
        return web.json_response(
            {
                "error": error,
                "stacktrace": stacktrace,
                "errcode": "plugin_reload_fail",
            },
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
        )

    @property
    def internal_server_error(self) -> web.Response:
        return web.json_response(
            {
                "error": "Internal server error",
                "errcode": "internal_server_error",
            },
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
        )

    @property
    def invalid_server(self) -> web.Response:
        return web.json_response(
            {
                "error": "Invalid registration server object in maubot configuration",
                "errcode": "invalid_server",
            },
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
        )

    @property
    def unsupported_plugin_loader(self) -> web.Response:
        return web.json_response(
            {
                "error": "Existing plugin with same ID uses unsupported plugin loader",
                "errcode": "unsupported_plugin_loader",
            },
            status=HTTPStatus.BAD_REQUEST,
        )

    @property
    def not_implemented(self) -> web.Response:
        return web.json_response(
            {
                "error": "Not implemented",
                "errcode": "not_implemented",
            },
            status=HTTPStatus.NOT_IMPLEMENTED,
        )

    @property
    def ok(self) -> web.Response:
        return web.json_response(
            {"success": True},
            status=HTTPStatus.OK,
        )

    @property
    def deleted(self) -> web.Response:
        return web.Response(status=HTTPStatus.NO_CONTENT)

    @staticmethod
    def found(data: dict) -> web.Response:
        return web.json_response(data, status=HTTPStatus.OK)

    @staticmethod
    def updated(data: dict, is_login: bool = False) -> web.Response:
        return web.json_response(data, status=HTTPStatus.ACCEPTED if is_login else HTTPStatus.OK)

    def logged_in(self, token: str) -> web.Response:
        return self.found({"token": token})

    def pong(self, user: str, features: dict) -> web.Response:
        return self.found({"username": user, "features": features})

    @staticmethod
    def created(data: dict) -> web.Response:
        return web.json_response(data, status=HTTPStatus.CREATED)


resp = _Response()



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

import logging

from attr import dataclass

from mautrix.client import SyncStore
from mautrix.types import FilterID, SyncToken, UserID
from mautrix.util.async_db import Connection, Database, UpgradeTable

upgrade_table = UpgradeTable(
    version_table_name="standalone_version", log=logging.getLogger("maubot.db")
)


@upgrade_table.register(description="Initial revision")
async def upgrade_v1(conn: Connection) -> None:
    await conn.execute(
        """CREATE TABLE IF NOT EXISTS standalone_next_batch (
            user_id    TEXT PRIMARY KEY,
            next_batch TEXT,
            filter_id  TEXT
        )"""
    )


find_q = "SELECT next_batch, filter_id FROM standalone_next_batch WHERE user_id=$1"
insert_q = "INSERT INTO standalone_next_batch (user_id, next_batch, filter_id) VALUES ($1, $2, $3)"
update_nb_q = "UPDATE standalone_next_batch SET next_batch=$1 WHERE user_id=$2"
update_filter_q = "UPDATE standalone_next_batch SET filter_id=$1 WHERE user_id=$2"


@dataclass
class NextBatch(SyncStore):
    db: Database
    user_id: UserID
    next_batch: SyncToken = ""
    filter_id: FilterID = ""

    async def load(self) -> NextBatch:
        row = await self.db.fetchrow(find_q, self.user_id)
        if row is not None:
            self.next_batch = row["next_batch"]
            self.filter_id = row["filter_id"]
        else:
            await self.db.execute(insert_q, self.user_id, self.next_batch, self.filter_id)
        return self

    async def put_filter_id(self, filter_id: FilterID) -> None:
        self.filter_id = filter_id
        await self.db.execute(update_filter_q, self.filter_id, self.user_id)

    async def put_next_batch(self, next_batch: SyncToken) -> None:
        self.next_batch = next_batch
        await self.db.execute(update_nb_q, self.next_batch, self.user_id)

    async def get_next_batch(self) -> SyncToken:
        return self.next_batch



from mautrix.util.async_db import Database

from .client import Client
from .instance import DatabaseEngine, Instance
from .upgrade import upgrade_table


def init(db: Database) -> None:
    for table in (Client, Instance):
        table.db = db


__all__ = ["upgrade_table", "init", "Client", "Instance", "DatabaseEngine"]



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from time import time
from zipfile import BadZipFile, ZipFile
import logging
import os
import sys

from packaging.version import Version
from ruamel.yaml import YAML, YAMLError

from mautrix.types import SerializerError

from ..__meta__ import __version__
from ..config import Config
from ..lib.zipimport import ZipImportError, zipimporter
from ..plugin_base import Plugin
from .abc import IDConflictError, PluginClass, PluginLoader
from .meta import DatabaseType, PluginMeta

current_version = Version(__version__)
yaml = YAML()


class MaubotZipImportError(Exception):
    pass


class MaubotZipMetaError(MaubotZipImportError):
    pass


class MaubotZipPreLoadError(MaubotZipImportError):
    pass


class MaubotZipLoadError(MaubotZipImportError):
    pass


class ZippedPluginLoader(PluginLoader):
    path_cache: dict[str, ZippedPluginLoader] = {}
    log: logging.Logger = logging.getLogger("maubot.loader.zip")
    trash_path: str = "delete"
    directories: list[str] = []

    path: str | None
    meta: PluginMeta | None
    main_class: str | None
    main_module: str | None
    _loaded: type[PluginClass] | None
    _importer: zipimporter | None
    _file: ZipFile | None

    def __init__(self, path: str) -> None:
        super().__init__()
        self.path = path
        self.meta = None
        self.main_class = None
        self.main_module = None
        self._loaded = None
        self._importer = None
        self._file = None
        self._load_meta()
        self._run_preload_checks(self._get_importer())
        try:
            existing = self.id_cache[self.meta.id]
            raise IDConflictError(
                f"Plugin with id {self.meta.id} already loaded from {existing.source}"
            )
        except KeyError:
            pass
        self.path_cache[self.path] = self
        self.id_cache[self.meta.id] = self
        self.log.debug(f"Preloaded plugin {self.meta.id} from {self.path}")

    def to_dict(self) -> dict:
        return {**super().to_dict(), "path": self.path}

    @classmethod
    def get(cls, path: str) -> ZippedPluginLoader:
        path = os.path.abspath(path)
        try:
            return cls.path_cache[path]
        except KeyError:
            return cls(path)

    @property
    def source(self) -> str:
        return self.path

    def __repr__(self) -> str:
        return (
            "<ZippedPlugin "
            f"path='{self.path}' "
            f"meta={self.meta} "
            f"loaded={self._loaded is not None}>"
        )

    def sync_read_file(self, path: str) -> bytes:
        return self._file.read(path)

    async def read_file(self, path: str) -> bytes:
        return self.sync_read_file(path)

    def sync_list_files(self, directory: str) -> list[str]:
        directory = directory.rstrip("/")
        return [
            file.filename
            for file in self._file.filelist
            if os.path.dirname(file.filename) == directory
        ]

    async def list_files(self, directory: str) -> list[str]:
        return self.sync_list_files(directory)

    @staticmethod
    def _read_meta(source) -> tuple[ZipFile, PluginMeta]:
        try:
            file = ZipFile(source)
            data = file.read("maubot.yaml")
        except FileNotFoundError as e:
            raise MaubotZipMetaError("Maubot plugin not found") from e
        except BadZipFile as e:
            raise MaubotZipMetaError("File is not a maubot plugin") from e
        except KeyError as e:
            raise MaubotZipMetaError("File does not contain a maubot plugin definition") from e
        try:
            meta_dict = yaml.load(data)
        except (YAMLError, KeyError, IndexError, ValueError) as e:
            raise MaubotZipMetaError("Maubot plugin definition file is not valid YAML") from e
        try:
            meta = PluginMeta.deserialize(meta_dict)
        except SerializerError as e:
            raise MaubotZipMetaError("Maubot plugin definition in file is invalid") from e
        if meta.maubot > current_version:
            raise MaubotZipMetaError(
                f"Plugin requires maubot {meta.maubot}, but this instance is {current_version}"
            )
        return file, meta

    @classmethod
    def verify_meta(cls, source) -> tuple[str, Version, DatabaseType | None]:
        _, meta = cls._read_meta(source)
        return meta.id, meta.version, meta.database_type if meta.database else None

    def _load_meta(self) -> None:
        file, meta = self._read_meta(self.path)
        if self.meta and meta.id != self.meta.id:
            raise MaubotZipMetaError("Maubot plugin ID changed during reload")
        self.meta = meta
        if "/" in meta.main_class:
            self.main_module, self.main_class = meta.main_class.split("/")[:2]
        else:
            self.main_module = meta.modules[-1]
            self.main_class = meta.main_class
        self._file = file

    def _get_importer(self, reset_cache: bool = False) -> zipimporter:
        try:
            if not self._importer or self._importer.archive != self.path:
                self._importer = zipimporter(self.path)
            if reset_cache:
                self._importer.reset_cache()
            return self._importer
        except ZipImportError as e:
            raise MaubotZipMetaError("File not found or not a maubot plugin") from e

    def _run_preload_checks(self, importer: zipimporter) -> None:
        try:
            code = importer.get_code(self.main_module.replace(".", "/"))
            if self.main_class not in code.co_names:
                raise MaubotZipPreLoadError(
                    f"Main class {self.main_class} not in {self.main_module}"
                )
        except ZipImportError as e:
            raise MaubotZipPreLoadError(f"Main module {self.main_module} not found in file") from e
        for module in self.meta.modules:
            try:
                importer.find_module(module)
            except ZipImportError as e:
                raise MaubotZipPreLoadError(f"Module {module} not found in file") from e

    async def load(self, reset_cache: bool = False) -> type[PluginClass]:
        try:
            return self._load(reset_cache)
        except MaubotZipImportError:
            self.log.exception(f"Failed to load {self.meta.id} v{self.meta.version}")
            raise

    def _load(self, reset_cache: bool = False) -> type[PluginClass]:
        if self._loaded is not None and not reset_cache:
            return self._loaded
        self._load_meta()
        importer = self._get_importer(reset_cache=reset_cache)
        self._run_preload_checks(importer)
        if reset_cache:
            self.log.debug(f"Re-preloaded plugin {self.meta.id} from {self.path}")
        for module in self.meta.modules:
            try:
                importer.load_module(module)
            except ZipImportError:
                raise MaubotZipLoadError(f"Module {module} not found in file")
            except Exception:
                raise MaubotZipLoadError(f"Failed to load module {module}")
        try:
            main_mod = sys.modules[self.main_module]
        except KeyError as e:
            raise MaubotZipLoadError(f"Main module {self.main_module} of plugin not found") from e
        try:
            plugin = getattr(main_mod, self.main_class)
        except AttributeError as e:
            raise MaubotZipLoadError(f"Main class {self.main_class} of plugin not found") from e
        if not issubclass(plugin, Plugin):
            raise MaubotZipLoadError("Main class of plugin does not extend maubot.Plugin")
        self._loaded = plugin
        self.log.debug(f"Loaded and imported plugin {self.meta.id} from {self.path}")
        return plugin

    async def reload(self, new_path: str | None = None) -> type[PluginClass]:
        self._unload()
        if new_path is not None and new_path != self.path:
            try:
                del self.path_cache[self.path]
            except KeyError:
                pass
            self.path = new_path
            self.path_cache[self.path] = self
        return await self.load(reset_cache=True)

    def _unload(self) -> None:
        for name, mod in list(sys.modules.items()):
            if (getattr(mod, "__file__", "") or "").startswith(self.path):
                del sys.modules[name]
        self._loaded = None
        self.log.debug(f"Unloaded plugin {self.meta.id} at {self.path}")

    async def delete(self) -> None:
        self._unload()
        try:
            del self.path_cache[self.path]
        except KeyError:
            pass
        try:
            del self.id_cache[self.meta.id]
        except KeyError:
            pass
        if self._importer:
            self._importer.remove_cache()
            self._importer = None
        self._loaded = None
        self.trash(self.path, reason="delete")
        self.meta = None
        self.path = None

    @classmethod
    def trash(cls, file_path: str, new_name: str | None = None, reason: str = "error") -> None:
        if cls.trash_path == "delete":
            try:
                os.remove(file_path)
            except FileNotFoundError:
                pass
        else:
            new_name = new_name or f"{int(time())}-{reason}-{os.path.basename(file_path)}"
            try:
                os.rename(file_path, os.path.abspath(os.path.join(cls.trash_path, new_name)))
            except OSError as e:
                cls.log.warning(f"Failed to rename {file_path}: {e} - trying to delete")
                try:
                    os.remove(file_path)
                except FileNotFoundError:
                    pass

    @classmethod
    def load_all(cls):
        cls.log.debug("Preloading plugins...")
        for directory in cls.directories:
            for file in os.listdir(directory):
                if not file.endswith(".mbp"):
                    continue
                path = os.path.abspath(os.path.join(directory, file))
                try:
                    cls.get(path)
                except MaubotZipImportError:
                    cls.log.exception(f"Failed to load plugin at {path}, trashing...")
                    cls.trash(path)
                except IDConflictError:
                    cls.log.error(f"Duplicate plugin ID at {path}, trashing...")
                    cls.trash(path)


def init(config: Config) -> None:
    ZippedPluginLoader.trash_path = config["plugin_directories.trash"]
    ZippedPluginLoader.directories = config["plugin_directories.load"]
    ZippedPluginLoader.load_all()



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import cast
import argparse
import asyncio
import copy
import importlib
import logging.config
import os.path
import signal
import sys

from aiohttp import ClientSession, hdrs, web
from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap
from yarl import URL

from mautrix.appservice import AppServiceServerMixin
from mautrix.client import SyncStream
from mautrix.types import (
    BaseMessageEventContentFuncs,
    Event,
    EventType,
    Filter,
    Membership,
    RoomEventFilter,
    RoomFilter,
    StrippedStateEvent,
)
from mautrix.util.async_db import Database, Scheme
from mautrix.util.config import BaseMissingError, RecursiveDict
from mautrix.util.logging import TraceLogger

from ..__meta__ import __version__
from ..loader import DatabaseType, PluginMeta
from ..matrix import MaubotMatrixClient
from ..plugin_base import Plugin
from ..plugin_server import PluginWebApp, PrefixResource
from ..server import AccessLogger
from .config import Config
from .database import NextBatch, upgrade_table
from .loader import FileSystemLoader

crypto_import_error = None

try:
    from mautrix.crypto import OlmMachine, PgCryptoStateStore, PgCryptoStore
except ImportError as err:
    crypto_import_error = err
    OlmMachine = PgCryptoStateStore = PgCryptoStore = None

parser = argparse.ArgumentParser(
    description="A plugin-based Matrix bot system -- standalone mode.",
    prog="python -m maubot.standalone",
)
parser.add_argument(
    "-c",
    "--config",
    type=str,
    default="config.yaml",
    metavar="<path>",
    help="the path to your config file",
)
parser.add_argument(
    "-b",
    "--base-config",
    type=str,
    default="pkg://maubot.standalone/example-config.yaml",
    metavar="<path>",
    help="the path to the example config (for automatic config updates)",
)
parser.add_argument(
    "-m",
    "--meta",
    type=str,
    default="maubot.yaml",
    metavar="<path>",
    help="the path to your plugin metadata file",
)
args = parser.parse_args()

config = Config(args.config, args.base_config)
config.load()
try:
    config.update()
except BaseMissingError:
    print("No example config found, not updating config")
except Exception as e:
    print("Failed to update config:", e)

logging.config.dictConfig(copy.deepcopy(config["logging"]))

log = logging.getLogger("maubot.init")

log.debug(f"Loading plugin metadata from {args.meta}")
yaml = YAML()
with open(args.meta, "r") as meta_file:
    meta: PluginMeta = PluginMeta.deserialize(yaml.load(meta_file.read()))

if "/" in meta.main_class:
    module, main_class = meta.main_class.split("/", 1)
else:
    module = meta.modules[-1]
    main_class = meta.main_class

if args.meta != "maubot.yaml" and os.path.dirname(args.meta) != "":
    sys.path.append(os.path.dirname(args.meta))
bot_module = importlib.import_module(module)
plugin: type[Plugin] = getattr(bot_module, main_class)
loader = FileSystemLoader(os.path.dirname(args.meta), meta)

log.info(f"Initializing standalone {meta.id} v{meta.version} on maubot {__version__}")

db = Database.create(
    config["database"],
    db_args=config.get("database_opts", {}),
    ignore_foreign_tables=True,
    log=logging.getLogger("maubot.db.standalone.upgrade"),
    upgrade_table=upgrade_table,
)

user_id = config["user.credentials.id"]
device_id = config["user.credentials.device_id"]
homeserver = config["user.credentials.homeserver"]
access_token = config["user.credentials.access_token"]
appservice_listener = config["user.appservice"]

crypto_store = state_store = None
if device_id and not OlmMachine:
    log.warning(
        "device_id set in config, but encryption dependencies not installed",
        exc_info=crypto_import_error,
    )
elif device_id:
    crypto_store = PgCryptoStore(account_id=user_id, pickle_key="mau.crypto", db=db)
    state_store = PgCryptoStateStore(db)

bot_config = None
if not meta.config and "base-config.yaml" in meta.extra_files:
    log.warning(
        "base-config.yaml in extra files, but config is not set to true. "
        "Assuming legacy plugin and loading config."
    )
    meta.config = True
if meta.config:
    log.debug("Loading config")
    config_class = plugin.get_config_class()

    def load() -> CommentedMap:
        return config["plugin_config"]

    def load_base() -> RecursiveDict[CommentedMap]:
        return RecursiveDict(config.load_base()["plugin_config"], CommentedMap)

    def save(data: RecursiveDict[CommentedMap]) -> None:
        config["plugin_config"] = data
        config.save()

    try:
        bot_config = config_class(load=load, load_base=load_base, save=save)
        bot_config.load_and_update()
    except Exception:
        log.fatal("Failed to load plugin config", exc_info=True)
        sys.exit(1)

if meta.webapp:
    web_app = web.Application()
    web_runner = web.AppRunner(web_app, access_log_class=AccessLogger)
    web_base_path = config["server.base_path"].rstrip("/")
    public_url = str(URL(config["server.public_url"]) / web_base_path.lstrip("/")).rstrip("/")
    plugin_webapp = PluginWebApp()

    async def _handle_plugin_request(req: web.Request) -> web.StreamResponse:
        if req.path.startswith(web_base_path):
            path_override = req.rel_url.path[len(web_base_path) :]
            url_override = req.rel_url.with_path(path_override).with_query(req.query_string)
            req = req.clone(rel_url=url_override)
            return await plugin_webapp.handle(req)
        return web.Response(status=404)

    resource = PrefixResource(web_base_path)
    resource.add_route(hdrs.METH_ANY, _handle_plugin_request)
    web_app.router.register_resource(resource)
elif appservice_listener:
    web_app = web.Application()
    web_runner = web.AppRunner(web_app, access_log_class=AccessLogger)
    public_url = plugin_webapp = None
else:
    web_app = web_runner = public_url = plugin_webapp = None

loop = asyncio.get_event_loop()

client: MaubotMatrixClient | None = None
bot: Plugin | None = None
appservice: AppServiceServerMixin | None = None


if appservice_listener:
    assert web_app is not None, "web_app is always set when appservice_listener is set"
    appservice = AppServiceServerMixin(
        ephemeral_events=True,
        encryption_events=True,
        log=logging.getLogger("maubot.appservice"),
        hs_token=config["user.hs_token"],
    )
    appservice.register_routes(web_app)

    @appservice.matrix_event_handler
    async def handle_appservice_event(evt: Event) -> None:
        if isinstance(evt.content, BaseMessageEventContentFuncs):
            evt.content.trim_reply_fallback()
        fake_sync_stream = SyncStream.JOINED_ROOM
        if evt.type.is_ephemeral:
            fake_sync_stream |= SyncStream.EPHEMERAL
        else:
            fake_sync_stream |= SyncStream.TIMELINE
        setattr(evt, "source", fake_sync_stream)
        tasks = client.dispatch_manual_event(evt.type, evt, include_global_handlers=True)
        await asyncio.gather(*tasks)


async def main():
    http_client = ClientSession(loop=loop)

    global client, bot

    await db.start()
    nb = await NextBatch(db, user_id).load()

    client_log = logging.getLogger("maubot.client").getChild(user_id)
    client = MaubotMatrixClient(
        mxid=user_id,
        base_url=homeserver,
        token=access_token,
        client_session=http_client,
        loop=loop,
        log=client_log,
        sync_store=nb,
        state_store=state_store,
        device_id=device_id,
    )
    if appservice:
        client.api.as_user_id = user_id
    client.ignore_first_sync = config["user.ignore_first_sync"]
    client.ignore_initial_sync = config["user.ignore_initial_sync"]
    if crypto_store:
        await crypto_store.upgrade_table.upgrade(db)
        await state_store.upgrade_table.upgrade(db)
        await crypto_store.open()

        client.crypto = OlmMachine(client, crypto_store, state_store)
        if appservice:
            appservice.otk_handler = client.crypto.handle_as_otk_counts
            appservice.device_list_handler = client.crypto.handle_as_device_lists
            appservice.to_device_handler = client.crypto.handle_as_to_device_event
            client.api.as_device_id = device_id
        crypto_device_id = await crypto_store.get_device_id()
        if crypto_device_id and crypto_device_id != device_id:
            log.fatal(
                "Mismatching device ID in crypto store and config "
                f"(store: {crypto_device_id}, config: {device_id})"
            )
            sys.exit(10)
        await client.crypto.load()
        if not crypto_device_id:
            await crypto_store.put_device_id(device_id)
        log.debug("Enabled encryption support")

    if web_runner:
        await web_runner.setup()
        site = web.TCPSite(web_runner, config["server.hostname"], config["server.port"])
        await site.start()
        log.info(f"Web server listening on {site.name}")

    while True:
        try:
            whoami = await client.whoami()
        except Exception as e:
            log.error(
                f"Failed to connect to homeserver: {type(e).__name__}: {e}"
                " - retrying in 10 seconds..."
            )
            await asyncio.sleep(10)
            continue
        if whoami.user_id != user_id:
            log.fatal(f"User ID mismatch: configured {user_id}, but server said {whoami.user_id}")
            sys.exit(11)
        elif whoami.device_id and device_id and whoami.device_id != device_id:
            log.fatal(
                f"Device ID mismatch: configured {device_id}, "
                f"but server said {whoami.device_id}"
            )
            sys.exit(12)
        log.debug(f"Confirmed connection as {whoami.user_id} / {whoami.device_id}")
        break

    if config["user.sync"]:
        if not nb.filter_id:
            filter_id = await client.create_filter(
                Filter(room=RoomFilter(timeline=RoomEventFilter(limit=50)))
            )
            await nb.put_filter_id(filter_id)
        _ = client.start(nb.filter_id)
    elif appservice_listener and crypto_store and not client.crypto.account.shared:
        await client.crypto.share_keys()

    if config["user.autojoin"]:
        log.debug("Autojoin is enabled")

        @client.on(EventType.ROOM_MEMBER)
        async def _handle_invite(evt: StrippedStateEvent) -> None:
            if evt.state_key == client.mxid and evt.content.membership == Membership.INVITE:
                await client.join_room(evt.room_id)

    displayname, avatar_url = config["user.displayname"], config["user.avatar_url"]
    if avatar_url != "disable":
        await client.set_avatar_url(avatar_url)
    if displayname != "disable":
        await client.set_displayname(displayname)

    plugin_log = cast(TraceLogger, logging.getLogger("maubot.instance.__main__"))
    if meta.database:
        if meta.database_type == DatabaseType.SQLALCHEMY:
            import sqlalchemy as sql

            plugin_db = sql.create_engine(config["database"])
            if db.scheme == Scheme.SQLITE:
                log.warning(
                    "Using SQLite with legacy plugins in standalone mode can cause database "
                    "locking issues. Switching to Postgres or updating the plugin to use the "
                    "new asyncpg/aiosqlite database interface is recommended."
                )
        elif meta.database_type == DatabaseType.ASYNCPG:
            plugin_db = db
            upgrade_table = plugin.get_db_upgrade_table()
            if upgrade_table:
                await upgrade_table.upgrade(plugin_db)
        else:
            log.fatal(f"Unsupported database type {meta.database_type}")
            sys.exit(13)
    else:
        plugin_db = None
    bot = plugin(
        client=client,
        loop=loop,
        http=http_client,
        instance_id="__main__",
        log=plugin_log,
        config=bot_config,
        database=plugin_db,
        webapp=plugin_webapp,
        webapp_url=public_url,
        loader=loader,
    )

    await bot.internal_start()


async def stop(suppress_stop_error: bool = False) -> None:
    if client:
        client.stop()
    if bot:
        try:
            await bot.internal_stop()
        except Exception:
            if not suppress_stop_error:
                log.exception("Error stopping bot")
    if web_runner and web_runner.server:
        try:
            await web_runner.shutdown()
            await web_runner.cleanup()
        except RuntimeError:
            if not suppress_stop_error:
                await db.stop()
                raise
    await db.stop()


signal.signal(signal.SIGINT, signal.default_int_handler)
signal.signal(signal.SIGTERM, signal.default_int_handler)


try:
    log.info("Starting plugin")
    loop.run_until_complete(main())
except SystemExit:
    loop.run_until_complete(stop(suppress_stop_error=True))
    loop.close()
    raise
except (Exception, KeyboardInterrupt) as e:
    if isinstance(e, KeyboardInterrupt):
        log.info("Startup interrupted, stopping")
    else:
        log.fatal("Failed to start plugin", exc_info=True)
    loop.run_until_complete(stop(suppress_stop_error=True))
    loop.close()
    sys.exit(1)

try:
    log.info("Startup completed, running forever")
    loop.run_forever()
except KeyboardInterrupt:
    log.info("Interrupt received, stopping")
    loop.run_until_complete(stop())
    loop.close()
    sys.exit(0)
except Exception:
    log.fatal("Fatal error in bot", exc_info=True)
    sys.exit(1)



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import IO
import json

from colorama import Fore
from yarl import URL
import aiohttp
import click

from ..cliq import cliq


class UploadError(Exception):
    pass


@cliq.command(help="Upload a maubot plugin")
@click.argument("path")
@click.option("-s", "--server", help="The maubot instance to upload the plugin to")
@cliq.with_authenticated_http
async def upload(path: str, server: str, sess: aiohttp.ClientSession) -> None:
    with open(path, "rb") as file:
        await upload_file(sess, file, server)


async def upload_file(sess: aiohttp.ClientSession, file: IO, server: str) -> None:
    url = (URL(server) / "_matrix/maubot/v1/plugins/upload").with_query({"allow_override": "true"})
    headers = {"Content-Type": "application/zip"}
    async with sess.post(url, data=file, headers=headers) as resp:
        if resp.status in (200, 201):
            data = await resp.json()
            print(
                f"{Fore.GREEN}Plugin {Fore.CYAN}{data['id']} v{data['version']}{Fore.GREEN} "
                f"uploaded to {Fore.CYAN}{server}{Fore.GREEN} successfully.{Fore.RESET}"
            )
        else:
            try:
                err = await resp.json()
                if "stacktrace" in err:
                    print(err["stacktrace"])
                err = err["error"]
            except (aiohttp.ContentTypeError, json.JSONDecodeError, KeyError):
                err = await resp.text()
            print(f"{Fore.RED}Failed to upload plugin: {err}{Fore.RESET}")



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import Any, Callable
import asyncio
import functools
import inspect
import traceback

from colorama import Fore
from prompt_toolkit.validation import Validator
from questionary import prompt
import aiohttp
import click

from ..base import app
from ..config import get_token
from .validators import ClickValidator, Required


def with_http(func):
    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        async with aiohttp.ClientSession() as sess:
            try:
                return await func(*args, sess=sess, **kwargs)
            except aiohttp.ClientError as e:
                print(f"{Fore.RED}Connection error: {e}{Fore.RESET}")

    return wrapper


def with_authenticated_http(func):
    @functools.wraps(func)
    async def wrapper(*args, server: str, **kwargs):
        server, token = get_token(server)
        if not token:
            return
        async with aiohttp.ClientSession(headers={"Authorization": f"Bearer {token}"}) as sess:
            try:
                return await func(*args, sess=sess, server=server, **kwargs)
            except aiohttp.ClientError as e:
                print(f"{Fore.RED}Connection error: {e}{Fore.RESET}")

    return wrapper


def command(help: str) -> Callable[[Callable], Callable]:
    def decorator(func) -> Callable:
        questions = getattr(func, "__inquirer_questions__", {}).copy()

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            for key, value in kwargs.items():
                if key not in questions:
                    continue
                if value is not None and (questions[key]["type"] != "confirm" or value != "null"):
                    questions.pop(key, None)
                try:
                    required_unless = questions[key].pop("required_unless")
                    if isinstance(required_unless, str) and kwargs[required_unless]:
                        questions.pop(key)
                    elif isinstance(required_unless, list):
                        for v in required_unless:
                            if kwargs[v]:
                                questions.pop(key)
                                break
                    elif isinstance(required_unless, dict):
                        for k, v in required_unless.items():
                            if kwargs.get(v, object()) == v:
                                questions.pop(key)
                                break
                except KeyError:
                    pass
            question_list = list(questions.values())
            question_list.reverse()
            resp = prompt(question_list, kbi_msg="Aborted!")
            if not resp and question_list:
                return
            kwargs = {**kwargs, **resp}

            try:
                res = func(*args, **kwargs)
                if inspect.isawaitable(res):
                    asyncio.run(res)
            except Exception:
                print(Fore.RED + "Fatal error running command" + Fore.RESET)
                traceback.print_exc()

        return app.command(help=help)(wrapper)

    return decorator


def yesno(val: str) -> bool | None:
    if not val:
        return None
    elif isinstance(val, bool):
        return val
    elif val.lower() in ("true", "t", "yes", "y"):
        return True
    elif val.lower() in ("false", "f", "no", "n"):
        return False


yesno.__name__ = "yes/no"


def option(
    short: str,
    long: str,
    message: str = None,
    help: str = None,
    click_type: str | Callable[[str], Any] = None,
    inq_type: str = None,
    validator: type[Validator] = None,
    required: bool = False,
    default: str | bool | None = None,
    is_flag: bool = False,
    prompt: bool = True,
    required_unless: str | list | dict = None,
) -> Callable[[Callable], Callable]:
    if not message:
        message = long[2].upper() + long[3:]

    if isinstance(validator, type) and issubclass(validator, ClickValidator):
        click_type = validator.click_type
    if is_flag:
        click_type = yesno

    def decorator(func) -> Callable:
        click.option(short, long, help=help, type=click_type)(func)
        if not prompt:
            return func
        if not hasattr(func, "__inquirer_questions__"):
            func.__inquirer_questions__ = {}
        q = {
            "type": (
                inq_type if isinstance(inq_type, str) else ("input" if not is_flag else "confirm")
            ),
            "name": long[2:],
            "message": message,
        }
        if required_unless is not None:
            q["required_unless"] = required_unless
        if default is not None:
            q["default"] = default
        if required or required_unless is not None:
            q["validate"] = Required(validator)
        elif validator:
            q["validate"] = validator
        func.__inquirer_questions__[long[2:]] = q
        return func

    return decorator



from typing import Any, Awaitable, Callable, Generator


class FutureAwaitable:
    def __init__(self, func: Callable[[], Awaitable[None]]) -> None:
        self._func = func

    def __await__(self) -> Generator[Any, None, None]:
        return self._func().__await__()



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import json
import webbrowser

from colorama import Fore
from yarl import URL
import aiohttp
import click

from ..cliq import cliq

history_count: int = 10

friendly_errors = {
    "server_not_found": (
        "Registration target server not found.\n\n"
        "To log in or register through maubot, you must add the server to the\n"
        "homeservers section in the config. If you only want to log in,\n"
        "leave the `secret` field empty."
    ),
    "registration_no_sso": (
        "The register operation is only for registering with a password.\n\n"
        "To register with SSO, simply leave out the --register flag."
    ),
}


async def list_servers(server: str, sess: aiohttp.ClientSession) -> None:
    url = URL(server) / "_matrix/maubot/v1/client/auth/servers"
    async with sess.get(url) as resp:
        data = await resp.json()
        print(f"{Fore.GREEN}Available Matrix servers for registration and login:{Fore.RESET}")
        for server in data.keys():
            print(f"* {Fore.CYAN}{server}{Fore.RESET}")


@cliq.command(help="Log into a Matrix account via the Maubot server")
@cliq.option("-h", "--homeserver", help="The homeserver to log into", required_unless="list")
@cliq.option(
    "-u", "--username", help="The username to log in with", required_unless=["list", "sso"]
)
@cliq.option(
    "-p",
    "--password",
    help="The password to log in with",
    inq_type="password",
    required_unless=["list", "sso"],
)
@cliq.option(
    "-s",
    "--server",
    help="The maubot instance to log in through",
    default="",
    required=False,
    prompt=False,
)
@click.option(
    "-r", "--register", help="Register instead of logging in", is_flag=True, default=False
)
@click.option(
    "-c",
    "--update-client",
    help="Instead of returning the access token, create or update a client in maubot using it",
    is_flag=True,
    default=False,
)
@click.option("-l", "--list", help="List available homeservers", is_flag=True, default=False)
@click.option(
    "-o", "--sso", help="Use single sign-on instead of password login", is_flag=True, default=False
)
@click.option(
    "-n",
    "--device-name",
    help="The initial e2ee device displayname (only for login)",
    default="Maubot",
    required=False,
)
@cliq.with_authenticated_http
async def auth(
    homeserver: str,
    username: str,
    password: str,
    server: str,
    register: bool,
    list: bool,
    update_client: bool,
    device_name: str,
    sso: bool,
    sess: aiohttp.ClientSession,
) -> None:
    if list:
        await list_servers(server, sess)
        return
    endpoint = "register" if register else "login"
    url = URL(server) / "_matrix/maubot/v1/client/auth" / homeserver / endpoint
    if update_client:
        url = url.update_query({"update_client": "true"})
    if sso:
        url = url.update_query({"sso": "true"})
        req_data = {"device_name": device_name}
    else:
        req_data = {"username": username, "password": password, "device_name": device_name}

    async with sess.post(url, json=req_data) as resp:
        if not 200 <= resp.status < 300:
            await print_error(resp, is_register=register)
        elif sso:
            await wait_sso(resp, sess, server, homeserver)
        else:
            await print_response(resp, is_register=register)


async def wait_sso(
    resp: aiohttp.ClientResponse, sess: aiohttp.ClientSession, server: str, homeserver: str
) -> None:
    data = await resp.json()
    sso_url, reg_id = data["sso_url"], data["id"]
    print(f"{Fore.GREEN}Opening {Fore.CYAN}{sso_url}{Fore.RESET}")
    webbrowser.open(sso_url, autoraise=True)
    print(f"{Fore.GREEN}Waiting for login token...{Fore.RESET}")
    wait_url = URL(server) / "_matrix/maubot/v1/client/auth" / homeserver / "sso" / reg_id / "wait"
    async with sess.post(wait_url, json={}) as resp:
        await print_response(resp, is_register=False)


async def print_response(resp: aiohttp.ClientResponse, is_register: bool) -> None:
    if resp.status == 200:
        data = await resp.json()
        action = "registered" if is_register else "logged in as"
        print(f"{Fore.GREEN}Successfully {action} {Fore.CYAN}{data['user_id']}{Fore.GREEN}.")
        print(f"{Fore.GREEN}Access token: {Fore.CYAN}{data['access_token']}{Fore.RESET}")
        print(f"{Fore.GREEN}Device ID: {Fore.CYAN}{data['device_id']}{Fore.RESET}")
    elif resp.status in (201, 202):
        data = await resp.json()
        action = "created" if resp.status == 201 else "updated"
        print(
            f"{Fore.GREEN}Successfully {action} client for "
            f"{Fore.CYAN}{data['id']}{Fore.GREEN} / "
            f"{Fore.CYAN}{data['device_id']}{Fore.GREEN}.{Fore.RESET}"
        )
    else:
        await print_error(resp, is_register)


async def print_error(resp: aiohttp.ClientResponse, is_register: bool) -> None:
    try:
        err_data = await resp.json()
        error = friendly_errors.get(err_data["errcode"], err_data["error"])
    except (aiohttp.ContentTypeError, json.JSONDecodeError, KeyError):
        error = await resp.text()
    action = "register" if is_register else "log in"
    print(f"{Fore.RED}Failed to {action}: {error}{Fore.RESET}")



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Awaitable, Callable
import base64
import logging

from aiohttp import web

from .auth import check_token
from .base import get_config
from .responses import resp

Handler = Callable[[web.Request], Awaitable[web.Response]]
log = logging.getLogger("maubot.server")


@web.middleware
async def auth(request: web.Request, handler: Handler) -> web.Response:
    subpath = request.path[len("/_matrix/maubot/v1") :]
    if (
        subpath.startswith("/auth/")
        or subpath.startswith("/client/auth_external_sso/complete/")
        or subpath == "/features"
        or subpath == "/logs"
    ):
        return await handler(request)
    err = check_token(request)
    if err is not None:
        return err
    return await handler(request)


@web.middleware
async def error(request: web.Request, handler: Handler) -> web.Response:
    try:
        return await handler(request)
    except web.HTTPException as ex:
        if ex.status_code == 404:
            return resp.path_not_found
        elif ex.status_code == 405:
            return resp.method_not_allowed
        return web.json_response(
            {
                "httpexception": {
                    "headers": {key: value for key, value in ex.headers.items()},
                    "class": type(ex).__name__,
                    "body": ex.text or base64.b64encode(ex.body),
                },
                "error": f"Unhandled HTTP {ex.status}: {ex.text[:128] or 'non-text response'}",
                "errcode": f"unhandled_http_{ex.status}",
            },
            status=ex.status,
        )
    except Exception:
        log.exception("Error in handler")
        return resp.internal_server_error


req_no = 0


def get_req_no():
    global req_no
    req_no += 1
    return req_no



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import List, Optional

from attr import dataclass
from packaging.version import InvalidVersion, Version

from mautrix.types import (
    ExtensibleEnum,
    SerializableAttrs,
    SerializerError,
    deserializer,
    serializer,
)

from ..__meta__ import __version__


@serializer(Version)
def serialize_version(version: Version) -> str:
    return str(version)


@deserializer(Version)
def deserialize_version(version: str) -> Version:
    try:
        return Version(version)
    except InvalidVersion as e:
        raise SerializerError("Invalid version") from e


class DatabaseType(ExtensibleEnum):
    SQLALCHEMY = "sqlalchemy"
    ASYNCPG = "asyncpg"


@dataclass
class PluginMeta(SerializableAttrs):
    id: str
    version: Version
    modules: List[str]
    main_class: str

    maubot: Version = Version(__version__)
    database: bool = False
    database_type: DatabaseType = DatabaseType.SQLALCHEMY
    config: bool = False
    webapp: bool = False
    license: str = ""
    extra_files: List[str] = []
    dependencies: List[str] = []
    soft_dependencies: List[str] = []

    @property
    def database_type_str(self) -> Optional[str]:
        return self.database_type.value if self.database else None



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import Callable
import asyncio
import json
import logging

from aiohttp import hdrs, web
from aiohttp.abc import AbstractAccessLogger
from yarl import URL
import pkg_resources

from mautrix.api import Method, PathBuilder

from .__meta__ import __version__
from .config import Config
from .plugin_server import PluginWebApp, PrefixResource


class AccessLogger(AbstractAccessLogger):
    def log(self, request: web.Request, response: web.Response, time: int):
        self.logger.info(
            f'{request.remote} "{request.method} {request.path} '
            f"{response.status} {response.body_length} "
            f'in {round(time, 4)}s"'
        )


class MaubotServer:
    log: logging.Logger = logging.getLogger("maubot.server")
    plugin_routes: dict[str, PluginWebApp]

    def __init__(
        self, management_api: web.Application, config: Config, loop: asyncio.AbstractEventLoop
    ) -> None:
        self.loop = loop or asyncio.get_event_loop()
        self.app = web.Application(loop=self.loop, client_max_size=100 * 1024 * 1024)
        self.config = config

        self.setup_appservice()
        self.app.add_subapp("/_matrix/maubot/v1", management_api)
        self.setup_instance_subapps()
        self.setup_management_ui()

        self.runner = web.AppRunner(self.app, access_log_class=AccessLogger)

    async def handle_plugin_path(self, request: web.Request) -> web.StreamResponse:
        for path, app in self.plugin_routes.items():
            if request.path.startswith(path):
                request = request.clone(
                    rel_url=request.rel_url.with_path(
                        request.rel_url.path[len(path) - 1 :]
                    ).with_query(request.query_string)
                )
                return await app.handle(request)
        return web.Response(status=404)

    def get_instance_subapp(self, instance_id: str) -> tuple[PluginWebApp, str]:
        subpath = self.config["server.plugin_base_path"] + instance_id + "/"
        url = self.config["server.public_url"] + subpath
        try:
            return self.plugin_routes[subpath], url
        except KeyError:
            app = PluginWebApp()
            self.plugin_routes[subpath] = app
            return app, url

    def remove_instance_webapp(self, instance_id: str) -> None:
        try:
            subpath = self.config["server.plugin_base_path"] + instance_id + "/"
            self.plugin_routes.pop(subpath).clear()
        except KeyError:
            return

    def setup_instance_subapps(self) -> None:
        self.plugin_routes = {}
        resource = PrefixResource(self.config["server.plugin_base_path"].rstrip("/"))
        resource.add_route(hdrs.METH_ANY, self.handle_plugin_path)
        self.app.router.register_resource(resource)

    def setup_appservice(self) -> None:
        as_path = PathBuilder("/_matrix/appservice/v1")
        self.add_route(Method.PUT, as_path.transactions, self.handle_transaction)

    def setup_management_ui(self) -> None:
        ui_base = self.config["server.ui_base_path"]
        if ui_base == "/":
            ui_base = ""
        directory = self.config[
            "server.override_resource_path"
        ] or pkg_resources.resource_filename("maubot", "management/frontend/build")
        self.app.router.add_static(f"{ui_base}/static", f"{directory}/static")
        self.setup_static_root_files(directory, ui_base)

        with open(f"{directory}/index.html", "r") as file:
            index_html = file.read()

        @web.middleware
        async def frontend_404_middleware(request: web.Request, handler) -> web.Response:
            if hasattr(handler, "__self__") and isinstance(handler.__self__, web.StaticResource):
                try:
                    return await handler(request)
                except web.HTTPNotFound:
                    return web.Response(body=index_html, content_type="text/html")
            return await handler(request)

        async def ui_base_redirect(_: web.Request) -> web.Response:
            raise web.HTTPFound(f"{ui_base}/")

        self.app.middlewares.append(frontend_404_middleware)
        self.app.router.add_get(
            f"{ui_base}/", lambda _: web.Response(body=index_html, content_type="text/html")
        )
        self.app.router.add_get(ui_base, ui_base_redirect)

    @staticmethod
    def _static_data(data: bytes, mime: str) -> Callable[[web.Request], web.Response]:
        def fn(_: web.Request) -> web.Response:
            return web.Response(body=data, content_type=mime)

        return fn

    def setup_static_root_files(self, directory: str, ui_base: str) -> None:
        files = {
            "asset-manifest.json": "application/json",
            "manifest.json": "application/json",
            "favicon.png": "image/png",
        }
        for file, mime in files.items():
            with open(f"{directory}/{file}", "rb") as stream:
                data = stream.read()
            self.app.router.add_get(f"{ui_base}/{file}", self._static_data(data, mime))

        public_url = self.config["server.public_url"]
        public_url_path = ""
        if public_url:
            public_url_path = URL(public_url).path.rstrip("/")

        api_path = f"{public_url_path}/_matrix/maubot/v1"

        path_prefix_response_body = json.dumps({"api_path": api_path.rstrip("/")})
        self.app.router.add_get(
            f"{ui_base}/paths.json",
            lambda _: web.Response(
                body=path_prefix_response_body, content_type="application/json"
            ),
        )

    def add_route(self, method: Method, path: PathBuilder, handler) -> None:
        self.app.router.add_route(method.value, str(path), handler)

    async def start(self) -> None:
        await self.runner.setup()
        site = web.TCPSite(self.runner, self.config["server.hostname"], self.config["server.port"])
        await site.start()
        self.log.info(f"Listening on {site.name}")

    async def stop(self) -> None:
        await self.runner.shutdown()
        await self.runner.cleanup()

    @staticmethod
    async def version(_: web.Request) -> web.Response:
        return web.json_response({"version": __version__})

    async def handle_transaction(self, request: web.Request) -> web.Response:
        return web.Response(status=501)



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import TYPE_CHECKING, Any, AsyncGenerator, cast
from collections import defaultdict
import asyncio
import inspect
import io
import logging
import os.path

from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap

from mautrix.types import UserID
from mautrix.util import background_task
from mautrix.util.async_db import Database, Scheme, UpgradeTable
from mautrix.util.async_getter_lock import async_getter_lock
from mautrix.util.config import BaseProxyConfig, RecursiveDict
from mautrix.util.logging import TraceLogger

from .client import Client
from .db import DatabaseEngine, Instance as DBInstance
from .lib.optionalalchemy import Engine, MetaData, create_engine
from .lib.plugin_db import ProxyPostgresDatabase
from .loader import DatabaseType, PluginLoader, ZippedPluginLoader
from .plugin_base import Plugin

if TYPE_CHECKING:
    from .__main__ import Maubot
    from .server import PluginWebApp

log: TraceLogger = cast(TraceLogger, logging.getLogger("maubot.instance"))
db_log: TraceLogger = cast(TraceLogger, logging.getLogger("maubot.instance_db"))

yaml = YAML()
yaml.indent(4)
yaml.width = 200


class PluginInstance(DBInstance):
    maubot: "Maubot" = None
    cache: dict[str, PluginInstance] = {}
    plugin_directories: list[str] = []
    _async_get_locks: dict[Any, asyncio.Lock] = defaultdict(lambda: asyncio.Lock())

    log: logging.Logger
    loader: PluginLoader | None
    client: Client | None
    plugin: Plugin | None
    config: BaseProxyConfig | None
    base_cfg: RecursiveDict[CommentedMap] | None
    base_cfg_str: str | None
    inst_db: sql.engine.Engine | Database | None
    inst_db_tables: dict | None
    inst_webapp: PluginWebApp | None
    inst_webapp_url: str | None
    started: bool

    def __init__(
        self,
        id: str,
        type: str,
        enabled: bool,
        primary_user: UserID,
        config: str = "",
        database_engine: DatabaseEngine | None = None,
    ) -> None:
        super().__init__(
            id=id,
            type=type,
            enabled=bool(enabled),
            primary_user=primary_user,
            config_str=config,
            database_engine=database_engine,
        )

    def __hash__(self) -> int:
        return hash(self.id)

    @classmethod
    def init_cls(cls, maubot: "Maubot") -> None:
        cls.maubot = maubot

    def postinit(self) -> None:
        self.log = log.getChild(self.id)
        self.cache[self.id] = self
        self.config = None
        self.started = False
        self.loader = None
        self.client = None
        self.plugin = None
        self.inst_db = None
        self.inst_db_tables = None
        self.inst_webapp = None
        self.inst_webapp_url = None
        self.base_cfg = None
        self.base_cfg_str = None

    def to_dict(self) -> dict:
        return {
            "id": self.id,
            "type": self.type,
            "enabled": self.enabled,
            "started": self.started,
            "primary_user": self.primary_user,
            "config": self.config_str,
            "base_config": self.base_cfg_str,
            "database": (
                self.inst_db is not None and self.maubot.config["api_features.instance_database"]
            ),
            "database_interface": self.loader.meta.database_type_str if self.loader else "unknown",
            "database_engine": self.database_engine_str,
        }

    def _introspect_sqlalchemy(self) -> dict:
        metadata = MetaData()
        metadata.reflect(self.inst_db)
        return {
            table.name: {
                "columns": {
                    column.name: {
                        "type": str(column.type),
                        "unique": column.unique or False,
                        "default": column.default,
                        "nullable": column.nullable,
                        "primary": column.primary_key,
                    }
                    for column in table.columns
                },
            }
            for table in metadata.tables.values()
        }

    async def _introspect_sqlite(self) -> dict:
        q = """
        SELECT
            m.name AS table_name,
            p.cid AS col_id,
            p.name AS column_name,
            p.type AS data_type,
            p.pk AS is_primary,
            p.dflt_value AS column_default,
            p.[notnull] AS is_nullable
        FROM sqlite_master m
        LEFT JOIN pragma_table_info((m.name)) p
        WHERE m.type = 'table'
        ORDER BY table_name, col_id
        """
        data = await self.inst_db.fetch(q)
        tables = defaultdict(lambda: {"columns": {}})
        for column in data:
            table_name = column["table_name"]
            col_name = column["column_name"]
            tables[table_name]["columns"][col_name] = {
                "type": column["data_type"],
                "nullable": bool(column["is_nullable"]),
                "default": column["column_default"],
                "primary": bool(column["is_primary"]),
                # TODO uniqueness?
            }
        return tables

    async def _introspect_postgres(self) -> dict:
        assert isinstance(self.inst_db, ProxyPostgresDatabase)
        q = """
        SELECT col.table_name, col.column_name, col.data_type, col.is_nullable, col.column_default,
               tc.constraint_type
        FROM information_schema.columns col
        LEFT JOIN information_schema.constraint_column_usage ccu
               ON ccu.column_name=col.column_name
        LEFT JOIN information_schema.table_constraints tc
               ON col.table_name=tc.table_name
              AND col.table_schema=tc.table_schema
              AND ccu.constraint_name=tc.constraint_name
              AND ccu.constraint_schema=tc.constraint_schema
              AND tc.constraint_type IN ('PRIMARY KEY', 'UNIQUE')
        WHERE col.table_schema=$1
        """
        data = await self.inst_db.fetch(q, self.inst_db.schema_name)
        tables = defaultdict(lambda: {"columns": {}})
        for column in data:
            table_name = column["table_name"]
            col_name = column["column_name"]
            tables[table_name]["columns"].setdefault(
                col_name,
                {
                    "type": column["data_type"],
                    "nullable": column["is_nullable"],
                    "default": column["column_default"],
                    "primary": False,
                    "unique": False,
                },
            )
            if column["constraint_type"] == "PRIMARY KEY":
                tables[table_name]["columns"][col_name]["primary"] = True
            elif column["constraint_type"] == "UNIQUE":
                tables[table_name]["columns"][col_name]["unique"] = True
        return tables

    async def get_db_tables(self) -> dict:
        if self.inst_db_tables is None:
            if isinstance(self.inst_db, Engine):
                self.inst_db_tables = self._introspect_sqlalchemy()
            elif self.inst_db.scheme == Scheme.SQLITE:
                self.inst_db_tables = await self._introspect_sqlite()
            else:
                self.inst_db_tables = await self._introspect_postgres()
        return self.inst_db_tables

    async def load(self) -> bool:
        if not self.loader:
            try:
                self.loader = PluginLoader.find(self.type)
            except KeyError:
                self.log.error(f"Failed to find loader for type {self.type}")
                await self.update_enabled(False)
                return False
        if not self.client:
            self.client = await Client.get(self.primary_user)
            if not self.client:
                self.log.error(f"Failed to get client for user {self.primary_user}")
                await self.update_enabled(False)
                return False
        if self.loader.meta.webapp:
            self.enable_webapp()
        self.log.debug("Plugin instance dependencies loaded")
        self.loader.references.add(self)
        self.client.references.add(self)
        return True

    def enable_webapp(self) -> None:
        self.inst_webapp, self.inst_webapp_url = self.maubot.server.get_instance_subapp(self.id)

    def disable_webapp(self) -> None:
        self.maubot.server.remove_instance_webapp(self.id)
        self.inst_webapp = None
        self.inst_webapp_url = None

    @property
    def _sqlite_db_path(self) -> str:
        return os.path.join(self.maubot.config["plugin_databases.sqlite"], f"{self.id}.db")

    async def delete(self) -> None:
        if self.loader is not None:
            self.loader.references.remove(self)
        if self.client is not None:
            self.client.references.remove(self)
        try:
            del self.cache[self.id]
        except KeyError:
            pass
        await super().delete()
        if self.inst_db:
            await self.stop_database()
            await self.delete_database()
        if self.inst_webapp:
            self.disable_webapp()

    def load_config(self) -> CommentedMap:
        return yaml.load(self.config_str)

    def save_config(self, data: RecursiveDict[CommentedMap]) -> None:
        buf = io.StringIO()
        yaml.dump(data, buf)
        val = buf.getvalue()
        if val != self.config_str:
            self.config_str = val
            self.log.debug("Creating background task to save updated config")
            background_task.create(self.update())

    async def start_database(
        self, upgrade_table: UpgradeTable | None = None, actually_start: bool = True
    ) -> None:
        if self.loader.meta.database_type == DatabaseType.SQLALCHEMY:
            if self.database_engine is None:
                await self.update_db_engine(DatabaseEngine.SQLITE)
            elif self.database_engine == DatabaseEngine.POSTGRES:
                raise RuntimeError(
                    "Instance database engine is marked as Postgres, but plugin uses legacy "
                    "database interface, which doesn't support postgres."
                )
            self.inst_db = create_engine(f"sqlite:///{self._sqlite_db_path}")
        elif self.loader.meta.database_type == DatabaseType.ASYNCPG:
            if self.database_engine is None:
                if os.path.exists(self._sqlite_db_path) or not self.maubot.plugin_postgres_db:
                    await self.update_db_engine(DatabaseEngine.SQLITE)
                else:
                    await self.update_db_engine(DatabaseEngine.POSTGRES)
            instance_db_log = db_log.getChild(self.id)
            if self.database_engine == DatabaseEngine.POSTGRES:
                if not self.maubot.plugin_postgres_db:
                    raise RuntimeError(
                        "Instance database engine is marked as Postgres, but this maubot isn't "
                        "configured to support Postgres for plugin databases"
                    )
                self.inst_db = ProxyPostgresDatabase(
                    pool=self.maubot.plugin_postgres_db,
                    instance_id=self.id,
                    max_conns=self.maubot.config["plugin_databases.postgres_max_conns_per_plugin"],
                    upgrade_table=upgrade_table,
                    log=instance_db_log,
                )
            else:
                self.inst_db = Database.create(
                    f"sqlite:{self._sqlite_db_path}",
                    upgrade_table=upgrade_table,
                    log=instance_db_log,
                )
            if actually_start:
                await self.inst_db.start()
        else:
            raise RuntimeError(f"Unrecognized database type {self.loader.meta.database_type}")

    async def stop_database(self) -> None:
        if isinstance(self.inst_db, Database):
            await self.inst_db.stop()
        elif isinstance(self.inst_db, Engine):
            self.inst_db.dispose()
        else:
            raise RuntimeError(f"Unknown database type {type(self.inst_db).__name__}")

    async def delete_database(self) -> None:
        if self.loader.meta.database_type == DatabaseType.SQLALCHEMY:
            ZippedPluginLoader.trash(self._sqlite_db_path, reason="deleted")
        elif self.loader.meta.database_type == DatabaseType.ASYNCPG:
            if self.inst_db is None:
                await self.start_database(None, actually_start=False)
            if isinstance(self.inst_db, ProxyPostgresDatabase):
                await self.inst_db.delete()
            else:
                ZippedPluginLoader.trash(self._sqlite_db_path, reason="deleted")
        else:
            raise RuntimeError(f"Unrecognized database type {self.loader.meta.database_type}")
        self.inst_db = None

    async def start(self) -> None:
        if self.started:
            self.log.warning("Ignoring start() call to already started plugin")
            return
        elif not self.enabled:
            self.log.warning("Plugin disabled, not starting.")
            return
        if not self.client or not self.loader:
            self.log.warning("Missing plugin instance dependencies, attempting to load...")
            if not await self.load():
                return
        cls = await self.loader.load()
        if self.loader.meta.webapp and self.inst_webapp is None:
            self.log.debug("Enabling webapp after plugin meta reload")
            self.enable_webapp()
        elif not self.loader.meta.webapp and self.inst_webapp is not None:
            self.log.debug("Disabling webapp after plugin meta reload")
            self.disable_webapp()
        if self.loader.meta.database:
            try:
                await self.start_database(cls.get_db_upgrade_table())
            except Exception:
                self.log.exception("Failed to start instance database")
                await self.update_enabled(False)
                return
        config_class = cls.get_config_class()
        if config_class:
            try:
                base = await self.loader.read_file("base-config.yaml")
                self.base_cfg = RecursiveDict(yaml.load(base.decode("utf-8")), CommentedMap)
                buf = io.StringIO()
                yaml.dump(self.base_cfg._data, buf)
                self.base_cfg_str = buf.getvalue()
            except (FileNotFoundError, KeyError):
                self.base_cfg = None
                self.base_cfg_str = None
            if self.base_cfg:
                base_cfg_func = self.base_cfg.clone
            else:

                def base_cfg_func() -> None:
                    return None

            self.config = config_class(self.load_config, base_cfg_func, self.save_config)
        self.plugin = cls(
            client=self.client.client,
            loop=self.maubot.loop,
            http=self.client.http_client,
            instance_id=self.id,
            log=self.log,
            config=self.config,
            database=self.inst_db,
            loader=self.loader,
            webapp=self.inst_webapp,
            webapp_url=self.inst_webapp_url,
        )
        try:
            await self.plugin.internal_start()
        except Exception:
            self.log.exception("Failed to start instance")
            await self.update_enabled(False)
            return
        self.started = True
        self.inst_db_tables = None
        self.log.info(
            f"Started instance of {self.loader.meta.id} v{self.loader.meta.version} "
            f"with user {self.client.id}"
        )

    async def stop(self) -> None:
        if not self.started:
            self.log.warning("Ignoring stop() call to non-running plugin")
            return
        self.log.debug("Stopping plugin instance...")
        self.started = False
        try:
            await self.plugin.internal_stop()
        except Exception:
            self.log.exception("Failed to stop instance")
        self.plugin = None
        if self.inst_db:
            try:
                await self.stop_database()
            except Exception:
                self.log.exception("Failed to stop instance database")
        self.inst_db_tables = None

    async def update_id(self, new_id: str | None) -> None:
        if new_id is not None and new_id.lower() != self.id:
            await super().update_id(new_id.lower())

    async def update_config(self, config: str | None) -> None:
        if config is None or self.config_str == config:
            return
        self.config_str = config
        if self.started and self.plugin is not None:
            res = self.plugin.on_external_config_update()
            if inspect.isawaitable(res):
                await res
        await self.update()

    async def update_primary_user(self, primary_user: UserID | None) -> bool:
        if primary_user is None or primary_user == self.primary_user:
            return True
        client = await Client.get(primary_user)
        if not client:
            return False
        await self.stop()
        self.primary_user = client.id
        if self.client:
            self.client.references.remove(self)
        self.client = client
        self.client.references.add(self)
        await self.update()
        await self.start()
        self.log.debug(f"Primary user switched to {self.client.id}")
        return True

    async def update_type(self, type: str | None) -> bool:
        if type is None or type == self.type:
            return True
        try:
            loader = PluginLoader.find(type)
        except KeyError:
            return False
        await self.stop()
        self.type = loader.meta.id
        if self.loader:
            self.loader.references.remove(self)
        self.loader = loader
        self.loader.references.add(self)
        await self.update()
        await self.start()
        self.log.debug(f"Type switched to {self.loader.meta.id}")
        return True

    async def update_started(self, started: bool) -> None:
        if started is not None and started != self.started:
            await (self.start() if started else self.stop())

    async def update_enabled(self, enabled: bool) -> None:
        if enabled is not None and enabled != self.enabled:
            self.enabled = enabled
            await self.update()

    async def update_db_engine(self, db_engine: DatabaseEngine | None) -> None:
        if db_engine is not None and db_engine != self.database_engine:
            self.database_engine = db_engine
            await self.update()

    @classmethod
    @async_getter_lock
    async def get(
        cls, instance_id: str, *, type: str | None = None, primary_user: UserID | None = None
    ) -> PluginInstance | None:
        try:
            return cls.cache[instance_id]
        except KeyError:
            pass

        instance = cast(cls, await super().get(instance_id))
        if instance is not None:
            instance.postinit()
            return instance

        if type and primary_user:
            instance = cls(instance_id, type=type, enabled=True, primary_user=primary_user)
            await instance.insert()
            instance.postinit()
            return instance

        return None

    @classmethod
    async def all(cls) -> AsyncGenerator[PluginInstance, None]:
        instances = await super().all()
        instance: PluginInstance
        for instance in instances:
            try:
                yield cls.cache[instance.id]
            except KeyError:
                instance.postinit()
                yield instance



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import os

from jinja2 import Template
from packaging.version import Version
from pkg_resources import resource_string

from .. import cliq
from ..cliq import SPDXValidator, VersionValidator
from ..util import spdx

loaded: bool = False
meta_template: Template
mod_template: Template
base_config: str


def load_templates():
    global mod_template, meta_template, base_config, loaded
    if loaded:
        return
    meta_template = Template(resource_string("maubot.cli", "res/maubot.yaml.j2").decode("utf-8"))
    mod_template = Template(resource_string("maubot.cli", "res/plugin.py.j2").decode("utf-8"))
    base_config = resource_string("maubot.cli", "res/config.yaml").decode("utf-8")
    loaded = True


@cliq.command(help="Initialize a new maubot plugin")
@cliq.option(
    "-n",
    "--name",
    help="The name of the project",
    required=True,
    default=os.path.basename(os.getcwd()),
)
@cliq.option(
    "-i",
    "--id",
    message="ID",
    required=True,
    help="The maubot plugin ID (Java package name format)",
)
@cliq.option(
    "-v",
    "--version",
    help="Initial version for project (PEP-440 format)",
    default="0.1.0",
    validator=VersionValidator,
    required=True,
)
@cliq.option(
    "-l",
    "--license",
    validator=SPDXValidator,
    default="AGPL-3.0-or-later",
    help="The license for the project (SPDX identifier)",
    required=False,
)
@cliq.option(
    "-c",
    "--config",
    message="Should the plugin include a config?",
    help="Include a config in the plugin stub",
    default=False,
    is_flag=True,
)
def init(name: str, id: str, version: Version, license: str, config: bool) -> None:
    load_templates()
    main_class = name[0].upper() + name[1:]
    meta = meta_template.render(
        id=id, version=str(version), license=license, config=config, main_class=main_class
    )
    with open("maubot.yaml", "w") as file:
        file.write(meta)
    if license:
        with open("LICENSE", "w") as file:
            file.write(spdx.get(license)["licenseText"])
    if not os.path.isdir(name):
        os.mkdir(name)
    mod = mod_template.render(config=config, name=main_class)
    with open(f"{name}/__init__.py", "w") as file:
        file.write(mod)
    if config:
        with open("base-config.yaml", "w") as file:
            file.write(base_config)



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import IO
from io import BytesIO
import asyncio
import glob
import os
import zipfile

from aiohttp import ClientSession
from colorama import Fore
from questionary import prompt
from ruamel.yaml import YAML, YAMLError
import click

from mautrix.types import SerializerError

from ...loader import PluginMeta
from ..base import app
from ..cliq import cliq
from ..cliq.validators import PathValidator
from ..config import get_token
from .upload import upload_file

yaml = YAML()


def zipdir(zip, dir):
    for root, dirs, files in os.walk(dir):
        for file in files:
            zip.write(os.path.join(root, file))


def read_meta(path: str) -> PluginMeta | None:
    try:
        with open(os.path.join(path, "maubot.yaml")) as meta_file:
            try:
                meta_dict = yaml.load(meta_file)
            except YAMLError as e:
                print(Fore.RED + "Failed to build plugin: Metadata file is not YAML")
                print(Fore.RED + str(e) + Fore.RESET)
                return None
    except FileNotFoundError:
        print(Fore.RED + "Failed to build plugin: Metadata file not found" + Fore.RESET)
        return None
    try:
        meta = PluginMeta.deserialize(meta_dict)
    except SerializerError as e:
        print(Fore.RED + "Failed to build plugin: Metadata file is not valid")
        print(Fore.RED + str(e) + Fore.RESET)
        return None
    return meta


def read_output_path(output: str, meta: PluginMeta) -> str | None:
    directory = os.getcwd()
    filename = f"{meta.id}-v{meta.version}.mbp"
    if not output:
        output = os.path.join(directory, filename)
    elif os.path.isdir(output):
        output = os.path.join(output, filename)
    elif os.path.exists(output):
        q = [{"type": "confirm", "name": "override", "message": f"{output} exists, override?"}]
        override = prompt(q)["override"]
        if not override:
            return None
        os.remove(output)
    return os.path.abspath(output)


def write_plugin(meta: PluginMeta, output: str | IO) -> None:
    with zipfile.ZipFile(output, "w") as zip:
        meta_dump = BytesIO()
        yaml.dump(meta.serialize(), meta_dump)
        zip.writestr("maubot.yaml", meta_dump.getvalue())

        for module in meta.modules:
            if os.path.isfile(f"{module}.py"):
                zip.write(f"{module}.py")
            elif module is not None and os.path.isdir(module):
                if os.path.isfile(f"{module}/__init__.py"):
                    zipdir(zip, module)
                else:
                    print(
                        Fore.YELLOW
                        + f"Module {module} is missing __init__.py, skipping"
                        + Fore.RESET
                    )
            else:
                print(Fore.YELLOW + f"Module {module} not found, skipping" + Fore.RESET)
        for pattern in meta.extra_files:
            for file in glob.iglob(pattern):
                zip.write(file)


@cliq.with_authenticated_http
async def upload_plugin(output: str | IO, *, server: str, sess: ClientSession) -> None:
    server, token = get_token(server)
    if not token:
        return
    if isinstance(output, str):
        with open(output, "rb") as file:
            await upload_file(sess, file, server)
    else:
        await upload_file(sess, output, server)


@app.command(
    short_help="Build a maubot plugin",
    help=(
        "Build a maubot plugin. First parameter is the path to root of the plugin "
        "to build. You can also use --output to specify output file."
    ),
)
@click.argument("path", default=os.getcwd())
@click.option(
    "-o", "--output", help="Path to output built plugin to", type=PathValidator.click_type
)
@click.option(
    "-u", "--upload", help="Upload plugin to server after building", is_flag=True, default=False
)
@click.option("-s", "--server", help="Server to upload built plugin to")
def build(path: str, output: str, upload: bool, server: str) -> None:
    meta = read_meta(path)
    if not meta:
        return
    if output or not upload:
        output = read_output_path(output, meta)
        if not output:
            return
    else:
        output = BytesIO()
    os.chdir(path)
    write_plugin(meta, output)
    if isinstance(output, str):
        print(f"{Fore.GREEN}Plugin built to {Fore.CYAN}{output}{Fore.GREEN}.{Fore.RESET}")
    else:
        output.seek(0)
    if upload:
        asyncio.run(upload_plugin(output, server=server))



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from datetime import datetime

from aiohttp import web
from asyncpg import PostgresError
import aiosqlite

from mautrix.util.async_db import Database

from ...instance import PluginInstance
from ...lib.optionalalchemy import Engine, IntegrityError, OperationalError, asc, desc
from .base import routes
from .responses import resp


@routes.get("/instance/{id}/database")
async def get_database(request: web.Request) -> web.Response:
    instance_id = request.match_info["id"].lower()
    instance = await PluginInstance.get(instance_id)
    if not instance:
        return resp.instance_not_found
    elif not instance.inst_db:
        return resp.plugin_has_no_database
    return web.json_response(await instance.get_db_tables())


@routes.get("/instance/{id}/database/{table}")
async def get_table(request: web.Request) -> web.Response:
    instance_id = request.match_info["id"].lower()
    instance = await PluginInstance.get(instance_id)
    if not instance:
        return resp.instance_not_found
    elif not instance.inst_db:
        return resp.plugin_has_no_database
    tables = await instance.get_db_tables()
    try:
        table = tables[request.match_info.get("table", "")]
    except KeyError:
        return resp.table_not_found
    try:
        order = [tuple(order.split(":")) for order in request.query.getall("order")]
        order = [
            (
                (asc if sort.lower() == "asc" else desc)(table.columns[column])
                if sort
                else table.columns[column]
            )
            for column, sort in order
        ]
    except KeyError:
        order = []
    limit = int(request.query.get("limit", "100"))
    if isinstance(instance.inst_db, Engine):
        return _execute_query_sqlalchemy(instance, table.select().order_by(*order).limit(limit))


@routes.post("/instance/{id}/database/query")
async def query(request: web.Request) -> web.Response:
    instance_id = request.match_info["id"].lower()
    instance = await PluginInstance.get(instance_id)
    if not instance:
        return resp.instance_not_found
    elif not instance.inst_db:
        return resp.plugin_has_no_database
    data = await request.json()
    try:
        sql_query = data["query"]
    except KeyError:
        return resp.query_missing
    rows_as_dict = data.get("rows_as_dict", False)
    if isinstance(instance.inst_db, Engine):
        return _execute_query_sqlalchemy(instance, sql_query, rows_as_dict)
    elif isinstance(instance.inst_db, Database):
        try:
            return await _execute_query_asyncpg(instance, sql_query, rows_as_dict)
        except (PostgresError, aiosqlite.Error) as e:
            return resp.sql_error(e, sql_query)
    else:
        return resp.unsupported_plugin_database


def check_type(val):
    if isinstance(val, datetime):
        return val.isoformat()
    return val


async def _execute_query_asyncpg(
    instance: PluginInstance, sql_query: str, rows_as_dict: bool = False
) -> web.Response:
    data = {"ok": True, "query": sql_query}
    if sql_query.upper().startswith("SELECT"):
        res = await instance.inst_db.fetch(sql_query)
        data["rows"] = [
            (
                {key: check_type(value) for key, value in row.items()}
                if rows_as_dict
                else [check_type(value) for value in row]
            )
            for row in res
        ]
        if len(res) > 0:
            # TODO can we find column names when there are no rows?
            data["columns"] = list(res[0].keys())
    else:
        res = await instance.inst_db.execute(sql_query)
        if isinstance(res, str):
            data["status_msg"] = res
        elif isinstance(res, aiosqlite.Cursor):
            data["rowcount"] = res.rowcount
            # data["inserted_primary_key"] = res.lastrowid
        else:
            data["status_msg"] = "unknown status"
    return web.json_response(data)


def _execute_query_sqlalchemy(
    instance: PluginInstance, sql_query: str, rows_as_dict: bool = False
) -> web.Response:
    assert isinstance(instance.inst_db, Engine)
    try:
        res = instance.inst_db.execute(sql_query)
    except IntegrityError as e:
        return resp.sql_integrity_error(e, sql_query)
    except OperationalError as e:
        return resp.sql_operational_error(e, sql_query)
    data = {
        "ok": True,
        "query": str(sql_query),
    }
    if res.returns_rows:
        data["rows"] = [
            (
                {key: check_type(value) for key, value in row.items()}
                if rows_as_dict
                else [check_type(value) for value in row]
            )
            for row in res
        ]
        data["columns"] = res.keys()
    else:
        data["rowcount"] = res.rowcount
    if res.is_insert:
        data["inserted_primary_key"] = res.inserted_primary_key
    return web.json_response(data)



import setuptools
import os

with open("requirements.txt") as reqs:
    install_requires = reqs.read().splitlines()

with open("optional-requirements.txt") as reqs:
    extras_require = {}
    current = []
    for line in reqs.read().splitlines():
        if line.startswith("#/"):
            extras_require[line[2:]] = current = []
        elif not line or line.startswith("#"):
            continue
        else:
            current.append(line)

extras_require["all"] = list({dep for deps in extras_require.values() for dep in deps})

path = os.path.join(os.path.abspath(os.path.dirname(__file__)), "maubot", "__meta__.py")
__version__ = "UNKNOWN"
with open(path) as f:
    exec(f.read())

setuptools.setup(
    name="maubot",
    version=__version__,
    url="https://github.com/maubot/maubot",
    project_urls={
        "Changelog": "https://github.com/maubot/maubot/blob/master/CHANGELOG.md",
    },

    author="Tulir Asokan",
    author_email="tulir@maunium.net",

    description="A plugin-based Matrix bot system.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",

    packages=setuptools.find_packages(),

    install_requires=install_requires,
    extras_require=extras_require,
    python_requires="~=3.10",

    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Topic :: Communications :: Chat",
        "Framework :: AsyncIO",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
    ],
    entry_points="""
        [console_scripts]
        mbc=maubot.cli:app
        [pytest11]
        maubot=maubot.testing
    """,
    data_files=[
        (".", ["maubot/example-config.yaml"]),
    ],
    package_data={
        "maubot": [
            "example-config.yaml",
            "management/frontend/build/*",
            "management/frontend/build/static/css/*",
            "management/frontend/build/static/js/*",
            "management/frontend/build/static/media/*",
            "py.typed",
        ],
        "maubot.cli": ["res/*"],
        "maubot.standalone": ["example-config.yaml"],
    },
)



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from mautrix.util.async_db import Connection, Scheme

from . import upgrade_table

legacy_version_query = "SELECT version_num FROM alembic_version"
last_legacy_version = "90aa88820eab"


@upgrade_table.register(description="Initial asyncpg revision")
async def upgrade_v1(conn: Connection, scheme: Scheme) -> None:
    if await conn.table_exists("alembic_version"):
        await migrate_legacy_to_v1(conn, scheme)
    else:
        return await create_v1_tables(conn)


async def create_v1_tables(conn: Connection) -> None:
    await conn.execute(
        """CREATE TABLE client (
            id           TEXT    PRIMARY KEY,
            homeserver   TEXT    NOT NULL,
            access_token TEXT    NOT NULL,
            device_id    TEXT    NOT NULL,
            enabled      BOOLEAN NOT NULL,

            next_batch TEXT NOT NULL,
            filter_id  TEXT NOT NULL,

            sync     BOOLEAN NOT NULL,
            autojoin BOOLEAN NOT NULL,
            online   BOOLEAN NOT NULL,

            displayname TEXT NOT NULL,
            avatar_url  TEXT NOT NULL
        )"""
    )
    await conn.execute(
        """CREATE TABLE instance (
            id           TEXT    PRIMARY KEY,
            type         TEXT    NOT NULL,
            enabled      BOOLEAN NOT NULL,
            primary_user TEXT    NOT NULL,
            config       TEXT    NOT NULL,
            FOREIGN KEY (primary_user) REFERENCES client(id) ON DELETE RESTRICT ON UPDATE CASCADE
        )"""
    )


async def migrate_legacy_to_v1(conn: Connection, scheme: Scheme) -> None:
    legacy_version = await conn.fetchval(legacy_version_query)
    if legacy_version != last_legacy_version:
        raise RuntimeError(
            "Legacy database is not on last version. "
            "Please upgrade the old database with alembic or drop it completely first."
        )
    await conn.execute("ALTER TABLE plugin RENAME TO instance")
    await update_state_store(conn, scheme)
    if scheme != Scheme.SQLITE:
        await varchar_to_text(conn)
    await conn.execute("DROP TABLE alembic_version")


async def update_state_store(conn: Connection, scheme: Scheme) -> None:
    # The Matrix state store already has more or less the correct schema, so set the version
    await conn.execute("CREATE TABLE mx_version (version INTEGER PRIMARY KEY)")
    await conn.execute("INSERT INTO mx_version (version) VALUES (2)")
    if scheme != Scheme.SQLITE:
        # Remove old uppercase membership type and recreate it as lowercase
        await conn.execute("ALTER TABLE mx_user_profile ALTER COLUMN membership TYPE TEXT")
        await conn.execute("DROP TYPE IF EXISTS membership")
        await conn.execute(
            "CREATE TYPE membership AS ENUM ('join', 'leave', 'invite', 'ban', 'knock')"
        )
        await conn.execute(
            "ALTER TABLE mx_user_profile ALTER COLUMN membership TYPE membership "
            "USING LOWER(membership)::membership"
        )
    else:
        # Recreate table to remove CHECK constraint and lowercase everything
        await conn.execute(
            """CREATE TABLE new_mx_user_profile (
                room_id     TEXT,
                user_id     TEXT,
                membership  TEXT NOT NULL
                            CHECK (membership IN ('join', 'leave', 'invite', 'ban', 'knock')),
                displayname TEXT,
                avatar_url  TEXT,
                PRIMARY KEY (room_id, user_id)
            )"""
        )
        await conn.execute(
            """
            INSERT INTO new_mx_user_profile (room_id, user_id, membership, displayname, avatar_url)
            SELECT room_id, user_id, LOWER(membership), displayname, avatar_url
            FROM mx_user_profile
            """
        )
        await conn.execute("DROP TABLE mx_user_profile")
        await conn.execute("ALTER TABLE new_mx_user_profile RENAME TO mx_user_profile")


async def varchar_to_text(conn: Connection) -> None:
    columns_to_adjust = {
        "client": (
            "id",
            "homeserver",
            "device_id",
            "next_batch",
            "filter_id",
            "displayname",
            "avatar_url",
        ),
        "instance": ("id", "type", "primary_user"),
        "mx_room_state": ("room_id",),
        "mx_user_profile": ("room_id", "user_id", "displayname", "avatar_url"),
    }
    for table, columns in columns_to_adjust.items():
        for column in columns:
            await conn.execute(f'ALTER TABLE "{table}" ALTER COLUMN {column} TYPE TEXT')



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

import asyncio
import sys

from mautrix.util.async_db import Database, DatabaseException, PostgresDatabase, Scheme
from mautrix.util.program import Program

from .__meta__ import __version__
from .client import Client
from .config import Config
from .db import init as init_db, upgrade_table
from .instance import PluginInstance
from .lib.future_awaitable import FutureAwaitable
from .lib.state_store import PgStateStore
from .loader.zip import init as init_zip_loader
from .management.api import init as init_mgmt_api
from .server import MaubotServer

try:
    from mautrix.crypto.store import PgCryptoStore
except ImportError:
    PgCryptoStore = None


class Maubot(Program):
    config: Config
    server: MaubotServer
    db: Database
    crypto_db: Database | None
    plugin_postgres_db: PostgresDatabase | None
    state_store: PgStateStore

    config_class = Config
    module = "maubot"
    name = "maubot"
    version = __version__
    command = "python -m maubot"
    description = "A plugin-based Matrix bot system."

    def prepare_log_websocket(self) -> None:
        from .management.api.log import init, stop_all

        init(self.loop)
        self.add_shutdown_actions(FutureAwaitable(stop_all))

    def prepare_arg_parser(self) -> None:
        super().prepare_arg_parser()
        self.parser.add_argument(
            "--ignore-unsupported-database",
            action="store_true",
            help="Run even if the database schema is too new",
        )
        self.parser.add_argument(
            "--ignore-foreign-tables",
            action="store_true",
            help="Run even if the database contains tables from other programs (like Synapse)",
        )

    def prepare_db(self) -> None:
        self.db = Database.create(
            self.config["database"],
            upgrade_table=upgrade_table,
            db_args=self.config["database_opts"],
            owner_name=self.name,
            ignore_foreign_tables=self.args.ignore_foreign_tables,
        )
        init_db(self.db)

        if PgCryptoStore:
            if self.config["crypto_database"] == "default":
                self.crypto_db = self.db
            else:
                self.crypto_db = Database.create(
                    self.config["crypto_database"],
                    upgrade_table=PgCryptoStore.upgrade_table,
                    ignore_foreign_tables=self.args.ignore_foreign_tables,
                )
        else:
            self.crypto_db = None

        if self.config["plugin_databases.postgres"] == "default":
            if self.db.scheme != Scheme.POSTGRES:
                self.log.critical(
                    'Using "default" as the postgres plugin database URL is only allowed if '
                    "the default database is postgres."
                )
                sys.exit(24)
            assert isinstance(self.db, PostgresDatabase)
            self.plugin_postgres_db = self.db
        elif self.config["plugin_databases.postgres"]:
            plugin_db = Database.create(
                self.config["plugin_databases.postgres"],
                db_args={
                    **self.config["database_opts"],
                    **self.config["plugin_databases.postgres_opts"],
                },
            )
            if plugin_db.scheme != Scheme.POSTGRES:
                self.log.critical("The plugin postgres database URL must be a postgres database")
                sys.exit(24)
            assert isinstance(plugin_db, PostgresDatabase)
            self.plugin_postgres_db = plugin_db
        else:
            self.plugin_postgres_db = None

    def prepare(self) -> None:
        super().prepare()

        if self.config["api_features.log"]:
            self.prepare_log_websocket()

        init_zip_loader(self.config)
        self.prepare_db()
        Client.init_cls(self)
        PluginInstance.init_cls(self)
        management_api = init_mgmt_api(self.config, self.loop)
        self.server = MaubotServer(management_api, self.config, self.loop)
        self.state_store = PgStateStore(self.db)

    async def start_db(self) -> None:
        self.log.debug("Starting database...")
        ignore_unsupported = self.args.ignore_unsupported_database
        self.db.upgrade_table.allow_unsupported = ignore_unsupported
        self.state_store.upgrade_table.allow_unsupported = ignore_unsupported
        try:
            await self.db.start()
            await self.state_store.upgrade_table.upgrade(self.db)
            if self.plugin_postgres_db and self.plugin_postgres_db is not self.db:
                await self.plugin_postgres_db.start()
            if self.crypto_db:
                PgCryptoStore.upgrade_table.allow_unsupported = ignore_unsupported
                if self.crypto_db is not self.db:
                    await self.crypto_db.start()
                else:
                    await PgCryptoStore.upgrade_table.upgrade(self.db)
        except DatabaseException as e:
            self.log.critical("Failed to initialize database", exc_info=e)
            if e.explanation:
                self.log.info(e.explanation)
            sys.exit(25)

    async def system_exit(self) -> None:
        if hasattr(self, "db"):
            self.log.trace("Stopping database due to SystemExit")
            await self.db.stop()

    async def start(self) -> None:
        await self.start_db()
        await asyncio.gather(*[plugin.load() async for plugin in PluginInstance.all()])
        await asyncio.gather(*[client.start() async for client in Client.all()])
        await super().start()
        async for plugin in PluginInstance.all():
            await plugin.load()
        await self.server.start()

    async def stop(self) -> None:
        self.add_shutdown_actions(*(client.stop() for client in Client.cache.values()))
        await super().stop()
        self.log.debug("Stopping server")
        try:
            await asyncio.wait_for(self.server.stop(), 5)
        except asyncio.TimeoutError:
            self.log.warning("Stopping server timed out")
        await self.db.stop()


Maubot().run()



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from time import time

from aiohttp import web

from mautrix.types import UserID
from mautrix.util.signed_token import sign_token, verify_token

from .base import get_config, routes
from .responses import resp


def is_valid_token(token: str) -> bool:
    data = verify_token(get_config()["server.unshared_secret"], token)
    if not data:
        return False
    return get_config().is_admin(data.get("user_id", None))


def create_token(user: UserID) -> str:
    return sign_token(
        get_config()["server.unshared_secret"],
        {
            "user_id": user,
            "created_at": int(time()),
        },
    )


def get_token(request: web.Request) -> str:
    token = request.headers.get("Authorization", "")
    if not token or not token.startswith("Bearer "):
        token = request.query.get("access_token", "")
    else:
        token = token[len("Bearer ") :]
    return token


def check_token(request: web.Request) -> web.Response | None:
    token = get_token(request)
    if not token:
        return resp.no_token
    elif not is_valid_token(token):
        return resp.invalid_token
    return None


@routes.post("/auth/ping")
async def ping(request: web.Request) -> web.Response:
    token = get_token(request)
    if not token:
        return resp.no_token

    data = verify_token(get_config()["server.unshared_secret"], token)
    if not data:
        return resp.invalid_token
    user = data.get("user_id", None)
    if not get_config().is_admin(user):
        return resp.invalid_token
    return resp.pong(user, get_config()["api_features"])
